Package: codelink
Version: 1.78.0
Title: Manipulation of Codelink microarray data
Author: Diego Diez
Maintainer: Diego Diez <diego10ruiz@gmail.com>
Depends: R (>= 2.10), BiocGenerics (>= 0.3.2), methods, Biobase (>=
        2.17.8), limma
Imports: annotate
Suggests: genefilter, parallel, knitr
LazyLoad: yes
Description: This package facilitates reading, preprocessing and manipulating
 Codelink microarray data. The raw data must be exported as text file using the
 Codelink software.
License: GPL-2
Collate: Codelink-class.R CodelinkSet-class.R file.R data.R norm.R
        plot.R CodelinkSet-methods.R CodelinkSet-tools.R
        CodelinkSetUnique-class.R CodelinkSetUnique-methods.R filter.R
        cluster.R
biocViews: Microarray, OneChannel, DataImport, Preprocessing
ByteCompile: yes
VignetteBuilder: knitr
URL: https://github.com/ddiez/codelink
BugReports: https://github.com/ddiez/codelink/issues
git_url: https://git.bioconductor.org/packages/codelink
git_branch: RELEASE_3_22
git_last_commit: d9a0b2d
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.22
Date/Publication: 2025-10-29
NeedsCompilation: no
Packaged: 2025-10-30 03:13:56 UTC; biocbuild

import(methods)

importFrom("grDevices", "colorRampPalette", "rainbow", "x11")
importFrom("graphics", "abline", "box", "image", "legend", "lines", "mtext", "par", "plot", "points")
importFrom("stats", "approx", "density", "lowess", "median", "sd")
importFrom("utils", "data", "read.table", "write.table")

importFrom(BiocGenerics, "annotation", "annotation<-", "normalize")
importFrom(annotate, "lookUp", "htmlpage")
importFrom(Biobase, 
		   "phenoData", "phenoData<-",
		   "featureData", "featureData<-",
		   "featureNames", "featureNames<-",
		   "experimentData", "experimentData<-",
		   "sampleNames", "sampleNames<-",
		   "assayDataElement", "assayDataElement<-",
		   "pData", "varMetadata", "exprs")
importFrom(limma,
		   "backgroundCorrect.matrix",
		   "normalizeCyclicLoess",
		   "normalizeMedianValues",
		   "normalizeQuantiles")

importClassesFrom(Biobase, "eSet", "ExpressionSet", "VersionedBiobase", "Versioned")

exportClasses(CodelinkSet, CodelinkSetUnique)
exportMethods(codCorrect, codNormalize, codPreprocess, probeNames, probeTypes,
	getSNR,	getInt, getBkg, getInfo, getWeight, getFlag, meanSNR, codPlot, exprs, chipDevice,
	filterSNR, codUniverse, writeCluster, writeCodelink, averageProbes)

S3method(as.matrix, Codelink)
S3method(dim, Codelink)

exportClasses(Codelink)
exportMethods("[", show, summary, summaryFlag, annotation)

export(readCodelink, writeCodelink, reportCodelink,
	plotMA, plotCorrelation, plotDensities, plotCV, imageCodelink,
	bkgdCorrect, normalize, mergeArray, createWeights, arrayNew,
	selCV, cutCV, fc2Cond, na2false, logCodelink,
	readCodelinkSet,Codelink2CodelinkSet)

# normalize()
# wrapper function to apply normalization methods to Codelink objects.
#normalize <- function(object, method = "quantiles", log.it = TRUE,
#	preserve = FALSE, weights = NULL, verbose = FALSE) {
setMethod("normalize", "Codelink", function(object, method="quantiles", log.it=TRUE, preserve=FALSE, weights=NULL, loess.method="fast") {
#normalize = function(object, method="quantiles", log.it=TRUE, preserve=FALSE, weights=NULL, loess.method="fast") {
	.Deprecated(msg="The Codelink interface is deprecated. Read Codelink data with 'readCodelinkSet' instead. More details in the vignette and documentation.")
	
	if(!is(object,"Codelink")) stop("A Codelink object is needed.")
	if(is.null(object$Ri)) stop("Background corrected intensities needed.")
	if(log.it && object$method$log) stop("Intensities already log2.")
	
	method <- match.arg(method, c("loess", "quantiles", "median"))
	
	if(!is.null(weights))
		if(is.matrix(weights))
			weights=apply(weights,1,min)
	
	object$Ni <- object$Ri
	if(log.it) object$Ni <- log2(object$Ni)
	switch(method,
				 loess = {
				 	#object$Ni <- normalize.loess(object$Ni, log.it = FALSE,
				 															 #weights = weights, verbose = verbose)
				 	object$Ni = normalizeCyclicLoess(object$Ni, weights=weights,method = loess.method)
				 	object$method$normalization <- "CyclicLoess"
				 },
				 quantiles = {
				 	object$Ni <- normalizeQuantiles(object$Ni)
				 	#object$Ni <- normalize.quantiles(object$Ni)
				 	object$method$normalization <- "quantiles"
				 },
				 median = {
				 	# no weights used right now.
				 	#object$Ni <- normalize.median(object$Ni,weights=weights)
				 	object$Ni <- normalizeMedianValues(object$Ni)
				 	object$method$normalization <- "median"
				 }
	)
	if(!preserve) object$Ri <- NULL
	if(log.it) object$method$log <- TRUE
	return(object)
})
#}

# normalize.median()
# based on limma implementation.
# normalize.median <- function(x, weights = NULL) {
# 	l <- dim(x)[2]
# 	
# 	if(is.null(weights)) {
# 		for(j in 1:l)
# 			x[, j] <- x[, j] - median(x[, j], na.rm = TRUE)
# 	} else {
# 		for(j in 1:l)
# 			x[, j] <- x[, j] - weighted.median(x[, j], weights[, j],
# 				na.rm = TRUE)
# 	}
# 
# 	x
# }


# normalize.loess()
# based on affy implementation.
# modified to allows missing values and weights.
# normalize.loess <- function (mat, 
# 	subset = sample(1:(dim(mat)[1]), min(c(5000, nrow(mat)))),
# 	epsilon = 10^-2, maxit = 1, log.it = TRUE,
# 	verbose = FALSE, span = 2/3, family.loess = "symmetric", weights = NULL) 
# {
#     J <- dim(mat)[2]
#     II <- dim(mat)[1]
#     newData <- mat
#     if (log.it) {
#         mat <- log2(mat)
#         newData <- log2(newData)
#     }
#     change <- epsilon + 1
#     fs <- matrix(0, II, J)
#     iter <- 0
# 
# 	if(is.null(weights))
#     	w <- c(0, rep(1, length(subset)), 0)
# 	else
# 		w <- weights
#     
# 	while (iter < maxit) {
#         iter <- iter + 1
#         means <- matrix(0, II, J)
#         for (j in 1:(J - 1)) {
#             for (k in (j + 1):J) {
#                 y <- newData[, j] - newData[, k]
#                 x <- (newData[, j] + newData[, k])/2
# 				
# 				# select genes that are not set to NA
# 				sel <- which(!is.na(as.character(y)))
# 				y <- y[sel]
# 				x <- x[sel]
# 				ww <- w[sel]
#                 
# 				index <- c(order(x)[1], subset, order(-x)[1])
#                 xx <- x[index]
#                 yy <- y[index]
# 				# reorder weights.
# 				ww <- ww[index]
# 
#                 aux <- loess(yy ~ xx, span = span, degree = 1, 
#                   weights = ww, family = family.loess, control = loess.control(surface = "direct"))
#                 aux <- predict(aux, data.frame(xx = x))/J
# 				# apply normalization to genes not NA.
#                 means[sel, j] <- means[sel, j] + aux
#                 means[sel, k] <- means[sel, k] - aux
#                 if (verbose) 
#                   cat(" + Done with", j, "vs", k, " in iteration ", 
#                     iter, "\n")
#             }
#         }
#         fs <- fs + means
#         newData <- mat - fs
#         change <- max(colMeans((means[subset, ])^2))
#         if (verbose) 
#             cat(iter, change, "\n")
#         oldfs <- fs
#     }
#     if ((change > epsilon) && (maxit > 1)) 
#         warning(paste("!! No convergence after", maxit, "iterations.\n"))
#     if (log.it) {
#         return(2^newData)
#     }
#     else return(newData)
# }
# 
# normalize.loess2 <- function(mat, ...) {
# 	J <- dim(mat)[2]
#     II <- dim(mat)[1]
# 
# 	means <- matrix(0, II, J)
# 	for (j in 1:(J - 1)) {
# 		for (k in (j + 1):J) {
# 			M <- mat[, j] - mat[, k]
# 			A <- (mat[, j] + mat[, k])/2
# 			tmp <- loessFit(M, A, ...)
# 			means[, j] <- means[, j] + tmp$fitted
# 			means[, k] <- means[, k] - tmp$fitted
# 		}
# 	}
# 	return(mat-means)
# }

#
setGeneric("writeCluster", function(x, file, chip, ...)
	standardGeneric("writeCluster"))
setMethod("writeCluster", "Codelink",
function(x, file, chip)
{
	mat <- as.matrix(x)
	samples  <- x$sample
	probes <- x$name

	writecluster(file, mat, samples, probes, chip)
})

setMethod("writeCluster", "CodelinkSet",
function(x, file, chip)
{
	mat <- exprs(x)
	samples <- sampleNames(x)
	probes <- probeNames(x)

	if(missing(chip))
		chip <- annotation(x)
	if(chip == "")
		stop("invalid chip name ''")
	
	writecluster(file, mat, samples, probes, chip)
})

setMethod("writeCluster", "MArrayLM", 
function(x, file, chip, probes)
{
	mat <- x$coef
	samples <- colnames(x$coef)
	# try to get probe names from MArrayLM.
	if(missing(probes))
		probes <- x$genes[, "ID"]
	
	writecluster(file, mat, samples, probes, chip)
})

setMethod("writeCluster", "ExpressionSet",
function(x, file, chip)
{
	mat <- exprs(x)
	samples <- sampleNames(x)
	probes <- featureNames(x)

	if(missing(chip))
		chip <- annotation(x)
	if(chip == "")
		stop("invalid chip name ''")
	
	writecluster(file, mat, samples, probes, chip)
})

writecluster <- function(file, mat, samples, probes, chip) {
	symbol <- lookUp(probes, chip, "SYMBOL")

	# make nice symbols?
	symbol[is.na(symbol)] <- ""
	symbol <- gsub("_predicted", "", symbol)
	
	info <- symbol

	header <- c("ORF", "NAME", samples)

	mat <- cbind(probes, info, mat)
	mat <- rbind(header, mat)
	
	write.table(mat, file = file, quote = FALSE, col.names = FALSE, row.names = FALSE, sep = "\t")
}

# readHeaderXLS()
# read header information from XLS exported file.
readHeaderXLS <- function(file, dec = FALSE) {
	nlines <- grep("Idx", scan(file, flush = TRUE, what = "",
		blank.lines.skip = F, quiet = TRUE))
	nlines <- nlines-1
	head <- list()
	if(!any(nlines)) stop("Not a Codelink XLS exported file.")
	head$header <- scan(file, nlines = nlines, sep = "\t", what = "",
		quiet = TRUE)
	head$nlines <- nlines
	if(any(foo <- grep("Product:", head$header)))
		head$product <- head$header[foo + 1] else head$product <- "Unknown"
	if(any(foo <- grep("Sample Name", head$header)))
		head$sample <- head$header[foo + 1] else head$sample <- file
	head$size <- arraySize(file, nlines)
	head$size <- head$size - 1
	if(dec) head$dec <- decDetect(file, nlines)
	head$columns <- scan(file, skip = nlines, nlines = 1, sep = "\t",
		what = "", quiet = TRUE)
	return(head)
}

# readHeader()
# read header information from codelink file.
readHeader <- function(file, dec = FALSE) {
	foo <- grep("-{80}", scan(file, nlines = 30, flush = T, quiet = TRUE,
		what = ""))
	if(!any(foo)) stop("Not a Codelink exported file.")
	nlines <- foo
	# Return list:
	head <- list(header = NULL, nlines = NULL, product = NULL, sample = NULL,
		size = NULL, dec = NULL, columns = NULL)
	head$header <- scan(file, nlines = nlines, sep = "\t", what = "",
		quiet = TRUE)
	head$nlines <- nlines
	if(any(foo <- grep("PRODUCT", head$header)))
		head$product <- head$header[foo + 1] else head$product <- "Unknown"
	if(any(foo <- grep("Sample Name", head$header)))
		head$sample <- head$header[foo + 1] else head$sample <- file
	head$size <- arraySize(file, nlines)
	if(dec) head$dec <- decDetect(file, nlines)
	head$columns <- scan(file, skip = nlines, nlines = 1, sep = "\t",
		what = "", quiet = TRUE)
	return(head)
}
# decDetect()
# detect decimal point.
decDetect <- function(file, nlines) {
	foo <- read.table(file, skip = nlines, nrows = 1, header = TRUE,
		sep = "\t", na.strings = "", comment.char = "", fill = TRUE)
        val <- NULL
        if(!is.null(foo$Spot_mean)) val <- foo$Spot_mean
        if(!is.null(foo$Raw_intensity) && is.null(val))
			val <- foo$Raw_intensity
        if(!is.null(foo$Normalized_intensity) && is.null(val))
			val <- foo$NormalNormalized_intensity
        if(is(val, "numeric")) dec <- "." else dec <- ","
	return(dec)
}
# arraySize()
# calculate chip size.
arraySize <- function(file, nlines) {
        data <- scan(file, skip = nlines + 1, sep = "\t", what = "integer",
			flush = TRUE, na.strings = "", quiet = TRUE)
        #ngenes <- length(data)  # number of genes.
		# Codelink exporter is buggy.
        ngenes <- length(data[!is.na(data)])
}
# checkColumns()
# check vector consistency.
checkColumns <- function(x, y)
{
	for(n in 1:length(x))
		if(x[n] != y[n]) return(FALSE)
	return(TRUE)
}

# .checkType()
# checks valid type with current file.
.checkType <- function(h, type, types) {
	any(h$columns %in% types[type])
}
# readCodelink()
# read of codelink data.
readCodelink = function(files=list.files(pattern = "TXT"), sample.name=NULL, flag, flag.weights, type.weights, dec=NULL, type="Spot", preserve=FALSE,	verbose=2, file.type="Codelink", check=TRUE, fix=FALSE, old=FALSE) {
	.Deprecated(msg="The Codelink interface is deprecated. Read Codelink data with 'readCodelinkSet' instead. More details in the vignette and documentation.")
	.readCodelinkRaw(files=files, sample.name=sample.name, flag=flag, flag.weights=flag.weights, type.weights=type.weights, dec=dec, type=type, preserve=preserve,	verbose=verbose, file.type=file.type, check=check, fix=fix, old=old)
}
.readCodelinkRaw <- function(files=list.files(pattern = "TXT"), sample.name=NULL, flag, flag.weights, type.weights, dec=NULL, type="Spot", preserve=FALSE,	verbose=2, file.type="Codelink", check=TRUE, fix=FALSE, old=FALSE)
{	
	if(length(files) == 0) stop("no Codelink files found.")
	if(!old)
		warning("'readCodelink' and 'readCodelinkSet' do not convert intensities to NA based on flags anymore, except for spots flagged as 'M' (MSR spot). Instead, createWeights() is used to assign weights. These weights can be used during normalization and linear modeling. To obtain the old behavior use parameter 'old=TRUE' (weights will be created anyway).")
	if(old)
		warning("calling 'readCodelink'/'readCodelinkSet' with 'old=TRUE'")
	
	nslides <- length(files)

	type <- match.arg(type,c("Spot", "Raw", "Norm"))
	types <- c("Spot" = "Spot_mean",
			   "Raw" = "Raw_intensity",
			   "Norm" = "Normalized_intensity")
	file.type <- match.arg(file.type, c("Codelink", "XLS"))

	if(file.type=="XLS") readHeader <- readHeaderXLS
	
	if(!is.null(sample.name) && (length(sample.name) != nslides))
		stop("sample.name must have equal length as files.")
	
	flags <- c("G", "L", "S", "C", "I", "M", "X")
	flag.cc.new <- list(M = NA) # only MSR spots are assigned NA.
	flag.cc <- list(M = NA, C = NA, I = NA)
	if(!missing(flag)) {
		for(k in names(flag)) {

			if(any(k %in% flags))
				flag.cc[[k]] <- flag[[k]]
			else
				warning("unknown flag ", k, " skipping.\n")
		}
	}
	
	flag.ww = c(
		"G"=1,
		"L"=1,
		"S"=1,
		"C"=0,
		"I"=0,
		"M"=0,
		"X"=0
		)
	if(!missing(flag.weights)) {
		for(k in names(flag.weights)) {
			if(any(names(flag.ww) == k))
				flag.ww[[k]]=flag.weights[[k]]
			else
				warning("unsupported weight flag", k, " (ignored...)")
		}
	}
	
	type.ww = c(
		"DISCOVERY"=1,
		"FIDUCIAL"=0,
		"POSITIVE"=0,
		"NEGATIVE"=0,
		"OTHER"=0
		)
	
	if(!missing(type.weights)) {
		for(k in names(type.weights)) {
			if(any(names(type.ww) == k))
				type.ww[[k]]=type.weights[[k]]
			else
				warning("unsupported weight type", k, " (ignored...)")
		}
	}
	
	# read Header.
	#head <- readHeader(files[1])

	# read arrays.
	for(n in 1:nslides) {
		if(verbose > 0 && n > 1) cat(paste("** reading file", n, "of",
			nslides, ":", files[n], "\n"))
		if(is.null(dec)) head <- readHeader(files[n], dec = TRUE)
		else head <- readHeader(files[n])
		
		if (!.checkType(head, type, types))
			stop(paste("File", files[n], "does not contain column", types[type], "(choose other type instead)"))

		if(n==1) {
			product <- head$product
			ngenes <- head$size
			if(verbose > 0) {
				cat("** product:", product, "\n")
				cat("** chip size:", ngenes, "\n")
				if(fix) cat("** requested probe rrder FIX\n")
					cat(paste("** reading file", n, "of", nslides, ":",
						files[n], "\n"))
			}

			# define Codelink object.
			Y <- matrix(NA, nrow = ngenes, ncol = nslides,
				dimnames = list(1:ngenes, 1:nslides))
			Z <- rep(NA, ngenes)
			X <- c(1:nslides)
			J <- matrix(NA, nrow = ngenes, ncol = 2,
				dimnames = list(1:ngenes, c("row","col")))
			
			method.list <- list(background = "NONE", normalization = "NONE",
				merge = "NONE", log = FALSE)
			head.list <- list(product = "NONE", sample = X,  file = X,
				name = Z, type = Z, id = Z, flag = Y, method = method.list,
				Bstdev = Y,	snr = Y, logical = J)

			switch(type,
				Spot = data.list <- list(Smean = Y, Bmedian = Y),
				Raw = data.list <- list(Ri = Y),
				Norm = data.list <- list(Ni = Y)
			)
			codelink <- c(head.list,  data.list)
		}

		if(verbose>2) print(head)
		if(verbose>1) cat(paste("  + detected", head$dec, 
			"as decimal symbol.\n"))

		if(head$product != product) stop("Different array type (",
			head$product, ")!")
		if(head$product == "Unknown") warning("Product type for ", files[n],
			" is unknown (missing PRODUCT field in header).")
		if(head$size != ngenes)	stop("Different number of probes (",
			head$size, ")\n")

		if(is.null(sample.name)) codelink$sample[n] <- head$sample
		if(verbose > 1) cat(paste("  + sample Name:", codelink$sample[n], "\n"))

		# read bulk data.
		#data <- read.table(files[n], skip=head$nlines, sep="\t", header=TRUE, row.names=1, quote="", dec=head$dec, comment.char="", nrows=head$size, blank.lines.skip=TRUE)
		data <- read.table(files[n], skip = head$nlines, sep = "\t",
			header = TRUE, quote = "", dec = head$dec, comment.char = "",
			nrows = head$size, blank.lines.skip = TRUE)

		#######################################################################
		# check data consistency.
		if(check && !fix) {
			if(n == 1) {
				if(file.type == "Codelink")
					check.first <- as.character(data[, "Probe_name"])
				else
					check.first <- as.character(data[, "Probe_Name"])
			} else {
				if(file.type == "Codelink")
					check.now <- as.character(data[, "Probe_name"])
				else
					check.now <- as.character(data[, "Probe_Name"])

				if(!checkColumns(check.first, check.now))
					warning("Different column order in files!\nCheck that files where generated with the same version of the Codelink Software.\nAlternatively you can turn on the fix argument and I will try to fix the order, using the Feature_id column if exists (optimal) or the Probe_name column otherwise (sub-optimal). Be aware that if I use the Probe_name, the control, fiducial and some few duplicated probes will probably be messed up. Depending on what you want to do, this would be unacceptable.")
			}
		}

		# try to fix data consistency.
		if(fix) {
			if(any(grep("Feature_id", head$column))) {
				cat("  + using Feature_id to fix order (optimal).\n")
				fix.col <- as.character(data[, "Feature_id"])
			} else {
				cat("  + using Probe_name to fix order (sub-optimal).\n")
				if(file.type == "Codelink")
					fix.col <- as.character(data[, "Probe_name"])
				 else
					fix.col <- as.character(data[, "Probe_Name"])
			}
			data <- data[order(fix.col), ]
		}
		#######################################################################

		# assign flag values.
		if(file.type == "Codelink") 
			codelink$flag[,n] <- as.character(data[, "Quality_flag"])
		else
			codelink$flag[,n] <- as.character(data[, "Quality.Flag"])
		# flag information.
		# allow combination of different flags.
		#flag.m <- grep("M", codelink$flag[,n])	# MSR masked spots.
		#flag.i <- grep("I", codelink$flag[,n])	# Irregular spots.
		#flag.c <- grep("C", codelink$flag[,n])	# Background contaminated
		#flag.s <- grep("S", codelink$flag[,n])	# Saturated spots.
		#flag.g <- grep("G", codelink$flag[,n])	# Good spots.
		#flag.l <- grep("L", codelink$flag[,n])	# Limit spots.
		#flag.x <- grep("X", codelink$flag[,n])	# User excluded spots.

		#if(verbose>1) {
			#cat("  + Quality flags:\n")
			#cat(paste("      G:",length(flag.g),"\t"))
			#cat(paste("      L:",length(flag.l),"\t"))
			#cat(paste("      M:",length(flag.m),"\t"))
			#cat(paste("      I:",length(flag.i),"\n"))
			#cat(paste("      C:",length(flag.c),"\t"))
			#cat(paste("      S:",length(flag.s),"\t"))
			#cat(paste("      X:",length(flag.x),"\n"))
		#}

		# assign intensity values.
		switch(type,
			Spot = {
				codelink$Smean[,n] <- data[, "Spot_mean"]
				codelink$Bmedian[,n] <- data[, "Bkgd_median"]

				# if found bkgd stdev get it to compute SNR.
				if(any(grep("Bkgd_stdev", head$columns)))
					codelink$Bstdev[,n] <- data[, "Bkgd_stdev"]
				else codelink$Bstdev[,n] <- NA

				# set values based on flags.
				#if(!is.null(flag$M)) {
					#codelink$Smean[flag.m, n] <- flag$M
					#codelink$Bmedian[flag.m, n] <- flag$M
				#}
				#if(!is.null(flag$I)) {
					#codelink$Smean[flag.i, n] <- flag$I
					#codelink$Bmedian[flag.i, n] <- flag$I
				#}
				#if(!is.null(flag$C)) {
					#codelink$Smean[flag.c, n] <- flag$C
					#codelink$Bmedian[flag.c, n] <- flag$C
				#}
				#if(!is.null(flag$S)) {
					#codelink$Smean[flag.s, n] <- flag$S
					#codelink$Bmedian[flag.s, n] <- flag$S
				#}
				#if(!is.null(flag$G)) {
					#codelink$Smean[flag.g, n] <- flag$G
					#codelink$Bmedian[flag.g, n] <- flag$G
				#}
				#if(!is.null(flag$L)) {
					#codelink$Smean[flag.l, n] <- flag$L
					#codelink$Bmedian[flag.l, n] <- flag$L
				#}
				#if(!is.null(flag$X)) {
					#codelink$Smean[flag.x, n] <- flag$X
					#codelink$Bmedian[flag.x, n] <- flag$X
				#}
			},
			Raw = {
				if(file.type == "Codelink")
					codelink$Ri[, n] <- data[, "Raw_intensity"]
				else
					codelink$Ri[, n] <- data[, "Raw_Intensity"]
				codelink$method$background <- "Codelink Subtract"
				# Set values based on Flags.
				#if(!is.null(flag$M)) {
					#codelink$Ri[flag.m, n] <- flag$M
				#}
				#if(!is.null(flag$I)) {
					#codelink$Ri[flag.i, n] <- flag$I
				#}
				#if(!is.null(flag$C)) {
					#codelink$Ri[flag.c, n] <- flag$C
				#}
				#if(!is.null(flag$S)) {
					#codelink$Ri[flag.s, n] <- flag$S
				#}
				#if(!is.null(flag$G)) {
					#codelink$Ri[flag.g, n] <- flag$G
				#}
				#if(!is.null(flag$L)) {
					#codelink$Ri[flag.l, n] <- flag$L
				#}
				#if(!is.null(flag$X)) {
					#codelink$Ri[flag.x, n] <- flag$X
				#}
			},
			Norm = {
				codelink$Ni[,n] <- data[, "Normalized_intensity"]
				codelink$method$background <- "Codelink Subtract"
				codelink$method$normalization <- "Codelink Median"
				# Set values based on Flags.
				#if(!is.null(flag$M)) {
					#codelink$Ni[flag.m, n] <- flag$M
				#}
				#if(!is.null(flag$I)) {
					#codelink$Ni[flag.i, n] <- flag$I
				#}
				#if(!is.null(flag$C)) {
					#codelink$Ni[flag.c, n] <- flag$C
				#}
				#if(!is.null(flag$S)) {
					#codelink$Ni[flag.s, n] <- flag$S
				#}
				#if(!is.null(flag$G)) {
					#codelink$Ni[flag.g, n] <- flag$G
				#}
				#if(!is.null(flag$L)) {
					#codelink$Ni[flag.l, n] <- flag$L
				#}
				#if(!is.null(flag$X)) {
					#codelink$Ni[flag.x, n] <- flag$X
				#}
			}
		)
		if(n == 1) {
			if(file.type == "Codelink") {
				codelink$name <- as.character(data[, "Probe_name"])
				codelink$type <- as.character(data[, "Probe_type"])
				if(!fix && !is.null(data$Feature_id))
					codelink$id <- as.character(data[, "Feature_id"])
				if(any(grep("Logical_row", head$columns)) &&
					any(grep("Logical_col", head$column))) {
					codelink$logical[,"row"] <- data[, "Logical_row"]
					codelink$logical[,"col"] <- data[, "Logical_col"]
				}
			} else {
				codelink$name <- as.character(data[, "Probe_Name"])
				codelink$type <- as.character(data[, "Type"])
			}
		}
	}
	
	codelink=new("Codelink", codelink)

	# fix flags.
	if(old) {
		cat("** applying flags ...")
		for(k in names(flag.cc)) {
			sel <- grep(k, codelink$flag)
			switch(type,
				   "Spot" = {
				   	codelink$Smean[sel] <- flag.cc[[k]]
				   	codelink$Bmedian[sel] <- flag.cc[[k]]
				   },
				   "Raw" = {
				   	codelink$Ri[sel] <- flag.cc[[k]]
				   },
				   "Norm" = {
				   	codelink$Ni[sel] <- flag.cc[[k]]
				   }
				   )
		}
		cat("OK\n")
	} else {
		cat("** applying flags to MSR spots ...")
		for(k in names(flag.cc.new)) {
			sel <- grep(k, codelink$flag)
			switch(type,
				   "Spot" = {
				   	codelink$Smean[sel] <- flag.cc.new[[k]]
				   	codelink$Bmedian[sel] <- flag.cc.new[[k]]
				   },
				   "Raw" = {
				   	codelink$Ri[sel] <- flag.cc.new[[k]]
				   },
				   "Norm" = {
				   	codelink$Ni[sel] <- flag.cc.new[[k]]
				   }
				   )
		}
		cat("OK\n")
	}
		
	cat("** computing weights ...")
	codelink$weight=createWeights(codelink,type.weights=type.ww,flag.weights=flag.ww)
	cat("OK\n")
	
	# compute SNR.
	if(type == "Spot") {
		if(verbose>0) cat("** computing SNR ...")
		codelink$snr <- SNR(codelink$Smean, codelink$Bmedian, codelink$Bstdev)
		if(verbose>0) cat("OK\n")
		if(!preserve) codelink$Bstdev <- NULL
	}
	
	if(!is.null(sample.name)) codelink$sample <- sample.name
	codelink$file <- files
	codelink$product <- product

	codelink
}
# writeCodelink()
# write Codelink object to file.
writeCodelink <- function(object, file, dec = ".", sep = "\t", flag = FALSE, chip) {
	if(!is(object, "Codelink")) stop("A Codelink object is needed.")

	write(paste("# BkgdCorrection method", object$method$background, sep = sep), file = file)
	write(paste("# Normalization method", object$method$normalization, sep = "\t"), file = file, append = TRUE)
	write(paste("# Log transformed", object$method$log, sep = "\t"), file=file, append = TRUE)

	if(missing(chip)) chip <- annotation(object)
	if(chip == "") stop("chip name is needed.")	
	probes <- object$name
	probes.acc <- lookUp(probes, chip, "ACCNUM", load = TRUE)
	probes.eg <- lookUp(probes, chip, "ENTREZID")

	val <- NULL
	if(!is.null(object$Smean)) val <- object$Smean
	if(!is.null(object$Ri)) val <- object$Ri
	if(!is.null(object$Ni)) val <- object$Ni

	tmp <- cbind(object$name, unlist(probes.acc), unlist(probes.eg), val, object$snr)
	head <- c("PROBE_NAME", "ACCESSION", "ENTREZID", paste("INTENSITY_", object$sample, sep = ""),
            paste("SNR_", object$sample, sep = ""))

	if(flag) {
		tmp <- cbind(tmp, object$flag)
		head <- c(head, paste("FLAG_", object$sample, sep =""))
	} 

	rownames(tmp) <- 1:dim(object)[1]
	tmp <- rbind(Index=head, tmp)
	write.table(tmp, file = file, quote = FALSE, sep = sep, dec = dec, col.names = FALSE)
}

# assignWeights=function(x,flag.w) {
# 	w = matrix(1, nrow = nrow(x), ncol = ncol(x))
# 	for(k in names(flag.w)) {
# 		sel=grep(k,x)
# 		# keep the lowest weight.
# 		w[sel][w[sel]>flag.w[k]]=flag.w[k]
# 	}
# 	w
# }

# "assignWeights<-"=function(object, weight) {
# 	assayDataElement(x, "weight")=weight
# }
# setGeneric("assignWeights")

setGeneric("createWeights", function(object,type.weights=NULL,flag.weights=NULL) standardGeneric("createWeights"))
setMethod("createWeights", "CodelinkSet",
function(object,type.weights=NULL,flag.weights=NULL) {
	w=createTypeWeights(object,type.weights)
	createFlagWeights(object,flag.weights,weights=w)
})
setMethod("createWeights", "Codelink",
function(object,type.weights=NULL,flag.weights=NULL) {
  w=createTypeWeights(object,type.weights)
  createFlagWeights(object,flag.weights,weights=w)
})


setGeneric("createFlagWeights", function(object,flag.weights=NULL,weights) standardGeneric("createFlagWeights"))
setMethod("createFlagWeights", "CodelinkSet",
function(object,flag.weights=NULL,weights) {
	if(missing(weights)) {
		w = array(1,dim(object))
		colnames(w)=sampleNames(object)
		rownames(w)=featureNames(object)	
	} else w = weights
	if(is.null(flag.weights))
		flag.weights = c("G"=1,"L"=1,"S"=1,"C"=0,"I"=0,"M"=0,"X"=0)
	for(k in names(flag.weights)) {
		sel=grep(k,getFlag(object))
		# keep the lowest weight.
		w[sel][w[sel]>flag.weights[k]]=flag.weights[k]
	}
	w
})

setMethod("createFlagWeights", "Codelink",
function(object,flag.weights=NULL,weights) {
  if(missing(weights)) {
  	w = array(1,dim(object))
  	colnames(w)=colnames(object$Bmedian)
  	rownames(w)=rownames(object$Bmedian)
  } else w = weights
  if(is.null(flag.weights))
  	flag.weights = c("G"=1,"L"=1,"S"=1,"C"=0,"I"=0,"M"=0,"X"=0)
  for(k in names(flag.weights)) {
  	sel=grep(k,object$flag)
  	# keep the lowest weight.
  	w[sel][w[sel]>flag.weights[k]]=flag.weights[k]
  }
  w
})

setGeneric("createTypeWeights", function(object,type.weights=NULL,weights) standardGeneric("createTypeWeights"))
setMethod("createTypeWeights","CodelinkSet",
function(object,type.weights=NULL,weights) {
	if(missing(weights)) {
		w = array(1,dim(object))
		colnames(w)=sampleNames(object)
		rownames(w)=featureNames(object)	
	} else w = weights
	if(is.null(type.weights))
		type.weights = c("DISCOVERY"=1,"FIDUCIAL"=0,"POSITIVE"=0,"NEGATIVE"=0, "OTHER"=0)
	for(k in names(type.weights)) {
		sel=grep(k,probeTypes(object))
		# keep the lowest weight.
		w[sel,][w[sel,]>type.weights[k]]=type.weights[k]
	}
	w
})

setMethod("createTypeWeights","Codelink",
function(object,type.weights=NULL,weights) {
  if(missing(weights)) {
  	w = array(1,dim(object))
  	colnames(w)=colnames(object$Bmedian)
  	rownames(w)=rownames(object$Bmedian)
  } else w = weights
  if(is.null(type.weights))
  	type.weights = c("DISCOVERY"=1,"FIDUCIAL"=0,"POSITIVE"=0,"NEGATIVE"=0, "OTHER"=0)
  for(k in names(type.weights)) {
  	sel=grep(k,object$type)
  	# keep the lowest weight.
  	w[sel,][w[sel,]>type.weights[k]]=type.weights[k]
  }
  w
})



# reportCodelink()
# report output to HTML.
reportCodelink <- function(object, chip, filename=NULL, title="Main title", probe.type=FALSE, other=NULL, other.ord=NULL) {
	if(!is(object,"Codelink") && !is(object,"character")) stop("Codelink object or character vector needed.")
	if(probe.type && !is(object,"Codelink")) stop("Codelink object needed putting type")
	if(is.null(filename)) stop("Filename needed.")

	if(is(object,"Codelink")) genes <- object$name else genes <- object
	if(!is.null(other)) {
		if(!is.null(other.ord)) {
			ord <- order(other[[other.ord]])

			genes <- genes[ord]
			for(n in names(other)) {
				other[[n]] <- other[[n]][ord]
			}
		}
	}
	if(probe.type) genes.type <- object$type
	genes.acc <- lookUp(genes, chip, "ACCNUM")
	genes.eg <- lookUp(genes, chip, "ENTREZID")
	genes.ug <- lookUp(genes, chip, "UNIGENE")
	genes.sym <- lookUp(genes, chip, "SYMBOL")
	genes.name <- lookUp(genes, chip, "GENENAME")

	# cleaup names.
	genes.sym[is.na(genes.sym)] <- ""
	genes.name[is.na(genes.name)] <- ""
	
	genes.list <- list(genes, genes.acc, as.character(genes.eg), as.character(genes.ug))
	head <- c("Id", "Genbank", "Entrez ID", "Unigene", "Symbol", "Name")
	other.list <- list(genes.sym, genes.name)

	if(probe.type) {
		head <- c(head, "TYPE")
		other.list <- c(other.list,list(genes.type))
	}
	if(!is.null(other)) {
		head <- c(head, names(other))
		other.list <- c(other.list, other)
	}
	htmlpage(genelist=genes.list, filename=filename, table.head=head, othernames = other.list, title=title, repository = list("gb","gb","ll","ug"))
	#css <- "<style> body {font-family: Arial, Helvetiva, sans-serif;} table {font-size: 8pt; border: solid 1px black;} th{color: white; background: black;} td{border: 0; max-width: 300px;vertical-align: top;}</style>"
	#cat(css, file=filename, append=TRUE)
}

# this is not yet used because of portability issues in the C code.
# readCodelinkFiles <- function(files = list.files(pattern = "TXT"),
# 	what = "Spot_mean", flag)
# {
# 	what <- match.arg(what, c("Spot_mean", "Raw_intensity", "Norm_intensity"))
# 	tmp <- .Call("R_read_codelink", files, what, PACKAGE="codelink")
# 
# 	flags <- c("G", "L", "S", "C", "I", "M", "X")
# 	flag.cc <- list(M=NA, C=NA, I=NA)
# 	if(!missing(flag)) {
# 		for(k in names(flag)) {
# 
# 			if(any(k %in% flags))
# 				flag.cc[[k]] <- flag[[k]]
# 			else
# 				warning("unknown flag ", k, " skipping.\n")
# 		}
# 	}
# 
# 	nfiles <- ncol(tmp$intensity)
# 	nprobes <- nrow(tmp$intensity)
# 	dims <- list(1:nprobes, 1:nfiles)
# 
# 	cod <- new("Codelink")
# 	cod$product <- tmp$product
# 	cod$file <- tmp$file
# 	cod$sample <- tmp$sample
# 	switch(tmp$what,
# 		Spot_mean = {
# 			cod$method$background = "NONE"
# 			cod$method$normalization = "NONE"
# 			cod$method$merge = "NONE"
# 			cod$method$log = FALSE
# 		},
# 		Raw_intensity = {
# 			cod$method$background = "Codelink subtract"
# 			cod$method$normalization = "NONE"
# 			cod$method$merge = "NONE"
# 			cod$method$log = FALSE
# 		},
# 		Norm_intensity = {
# 			cod$method$background = "Codelink subtract"
# 			cod$method$normalization = "Codelink median"
# 			cod$method$merge = "NONE"
# 			cod$method$log = FALSE
# 		}
# 	)
# 	cod$name <- tmp$name
# 	cod$type <- tmp$type
# 	cod$id <- tmp$id
# 	cod$logical <- cbind(row = tmp$row, col = tmp$col)
# 	cod$flag <- tmp$flag
# 	cod$Smean <- tmp$intensity
# 	cod$Bmedian <- tmp$background
# 	cod$Bstdev <- tmp$bstdev
# 	
# 	# fix flags.
# 	cat("** applying flags ...")
# 	for(k in names(flag.cc)) {
# 		sel <- grep(k, cod$flag)
# 		cod$Smean[sel] <- flag.cc[[k]]
# 		cod$Bmedian[sel] <- flag.cc[[k]]
# 	}
# 	cat("OK\n")
# 	
# 	# compute snr.
# 	cat("** computing SNR ...")
# 	cod$snr <- SNR(tmp$intensity, tmp$background, tmp$bstdev)
# 	cat("OK\n")
# 
# 	# fix dimnames.
# 	dimnames(cod$flag) <- dims
# 	dimnames(cod$snr) <- dims
# 	dimnames(cod$Smean) <- dims
# 	dimnames(cod$Bmedian) <- dims
# 	dimnames(cod$Bstdev) <- dims
# 
# 	cod
#}

# mergeArray()
# returns a codelink data.frame mergin chips attending a function
# like median, mean, etc... default mean.
mergeArray <- function(object, class, names=NULL, method="mean",
			log.it=FALSE, merge.snr=TRUE) {
	if(!is(object,"Codelink")) stop("A Codelink object needed.")
	if(is.null(object$Ni) && is.null(object$Ri)) stop("Ni or Ri slots needed.")
	if(is.null(names)) stop("Group names needed.")
	if(log.it && object$method$log) stop("Data already log transformed.")
	if(is.null(object$snr) || !merge.snr) doSNR <- FALSE else doSNR <- TRUE

	l <- levels(as.factor(class))
	if(length(names)!=length(l))
		stop("Number of classes and class names disagree.")

	method = match.arg(method,c("mean"))
	dimx <- dim(object)[1]
	dimy <- dim(object)[2]
	if(!is.null(object$Ri)) data <- object$Ri else data <- object$Ni
	val <- matrix(nrow=dimx, ncol=length(l), dimnames=list(rownames(data),l))
	val.cv <- matrix(nrow=dimx, ncol=length(l), dimnames=list(rownames(data),l))
	if(doSNR) snr <- matrix(nrow=dimx, ncol=length(l), dimnames=list(rownames(data),l))
	cat("\n\n")
	switch(method,
		mean = {
			d <- 0
			for(n in l) {
				d <- d+1
				sel <- class==n
				cat("  Merging: ",object$sample[sel],"as", names[d],"\n")
				val[,as.numeric(n)] <- apply(data, 1, function(x) mean(x[sel], na.rm=TRUE))
				val.cv[,as.numeric(n)] <- apply(data, 1, function(x) sd(x[sel], na.rm=TRUE)/mean(x[sel],na.rm=TRUE))
				if(doSNR) snr[,as.numeric(n)] <- apply(object$snr, 1, function(x) mean(x[sel], na.rm=TRUE))
			}
			object$method$merge <- "mean"
		}
	)
	cat("\n")
	if(log.it) {
		if(!is.null(object$Ri)) object$Ri <- log2(val) else object$Ni <- log2(val)
		object$method$log <- TRUE
	} else {
		if(!is.null(object$Ri)) object$Ri <- val else object$Ni <- val
	}
	object$cv <- val.cv
	if(doSNR) object$snr <- snr
	object$sample <- names
	return(object)
}
## bkgdCorrect()
# Correct Spot intensity by background.
bkgdCorrect <- function(object, method = "half", preserve = FALSE,
                        verbose = FALSE, offset = 0) {
	.Deprecated(msg="The Codelink interface is deprecated. Read Codelink data with 'readCodelinkSet' instead. More details in the vignette and documentation.")
	
	if(!is(object,"Codelink")) stop("Codelink object needed.")
	method <- match.arg(method, c("none", "subtract", "half", "normexp"))
	
	newInt=backgroundCorrect.matrix(object$Smean, object$Bmedian, method=method, offset=offset)
	object$Ri=newInt
	object$method$background=method
	
	#switch(method,
	#	none = {
	#		#object$Ri <- object$Smean
	#		object$method$background <- "NONE"
	#	},
	#	subtract = {
	#		#object$Ri <- object$Smean - object$Bmedian
	#		object$method$background <- "subtract"
	#	},
	#	half = {
	#		#object$Ri <- pmax(object$Smean - object$Bmedian, 0.5)
	#		object$method$background <- "half"
	#	},
	#	normexp = {
	#		object$Ri <- object$Smean
	#		for (j in 1:ncol(object$Smean)) {
	#		    x <- object$Smean[, j] - object$Bmedian[, j]
	#		    out <- normexp.fit(x)
	#		    object$Ri[, j] <- normexp.signal(out$par, x)
	#		    if (verbose)
	#				cat(" + Corrected array", j, "\n")
	#		}
	#		object$method$background <- "normexp"
	#	}
	#)
	if(!preserve) object$Smean <- NULL
	if(!preserve) object$Bmedian <- NULL
	if(offset) object$Ri <- object$Ri + offset
	return(object)
}

# bgcorrect.normexp()
# bgcorrect.normexp <- function(x, y, verbose = FALSE) {
# 	#z <- matrix(NA, nrow = nrow(x), ncol = ncol(x), dimnames = dimnames(x))
# 	for (j in 1:ncol(x)) {
# 		xc <- x[, j] - y[, j]
# 		out <- normexp.fit(xc)
# 		x[, j] <- normexp.signal(out$par, xc)
# 		if (verbose)
# 			cat(" + Corrected array", j, "\n")
# 	}
# 	x
# }

## logCodelink()
# apply log2 to Codelink object.
logCodelink <- function(object) {
	if(!is(object,"Codelink")) stop("Codelink object needed.")
	if(object$method$log) stop("Already log2 transformed.")

	if(!is.null(object$Smean)) object$Smean <- log2(object$Smean)
	if(!is.null(object$Ri)) object$Ri <- log2(object$Ri)
	if(!is.null(object$Ni)) object$Ni <- log2(object$Ni)

	object$method$log <- TRUE
	return(object)
}

## SNR()
# Compute Signal to Noise Ratio.
SNR <- function(Smean, Bmedian, Bstdev) {
	Smean / (Bmedian + 1.5 * Bstdev)
}
## fc2Cond()
# rewrite of fc to be more intuitive:
fc2Cond <- function(object, cond1=NULL, cond2=NULL, fc=1.0, verbose=FALSE) {
	if(!is(object, "Codelink")) stop("Codelink object needed.")
	if(!is(cond1, "numeric") && !is(cond2, "numeric")) {
		#cat("Cond1:", cond1,"\n")
		#cat("Cond2:", cond2,"\n")
		cond1.inx <- which(object$sample==cond1)
		cond2.inx <- which(object$sample==cond2)
		#cat("Index1:", cond1.inx, "\n")
		#cat("Index2:", cond2.inx, "\n")
		if(!any(cond1.inx)) stop("Unknown sample ",cond1)
		if(!any(cond2.inx)) stop("Unknown sample ",cond2)
		if(length(cond1.inx)>1) stop(cond1," matches more than one condition in sample slot.")
		if(length(cond2.inx)>1) stop(cond2," matches more than one condition in sample slot.")
	} else {
		cond1.inx <- cond1
		cond2.inx <- cond2
	}
	if(!object$method$log) {
		if(verbose) cat("Commputing log2...\n")
		object <- logCodelink(object)
	}
	if(verbose) cat("Computing FC for", cond1,"-",cond2,"...\n")
	fc.0 <- abs(object$Ni[,cond1.inx] - object$Ni[, cond2.inx])
	return(na2false(fc.0 >= fc))
}
## fc()
# Select denes based in fold change between two conditions.
#fc <- function(object, cond1=NULL, cond2=NULL, fc=1.0) {
#	if(!is(object, "Codelink")) stop("Codelink object needed.")
#	if(cond1 > dim(object)[2] || cond2 > dim(object)[2]) stop("Invalid conditions.")
#	if(cond1 == cond2) stop("Fold changes along same conditions are equal to 0.")
#	if(is.null(object$Ni)) stop("Normalize data first.")
#	if(object$method$log) {
#		fc.0 <- abs(object$Ni[,cond1] - object$Ni[,cond2])
#	} else {
#		stop("Only logged data at this moment supported.")
#	}
#	return(na2false(fc.0 >= fc))
#}

## cutCV()
# Calculate cutoff based on C.V.
cutCV <- function(object, subset=c(1:dim(object)[2])) {
	if(!is(object, "Codelink")) stop("Codelink object needed.")
	if(object$method$merge == "NONE") stop("Merged object needed.")
	cut.mean <- mean(apply(data$cv[,subset],2,function(x) median(x,na.rm=TRUE)), na.rm=TRUE)
	cut.sd <- sd(apply(data$cv[,subset],2,function(x) median(x,na.rm=TRUE)), na.rm=TRUE)
	cut <- cut.mean + 3*cut.sd
	return(cut)
}
## selCV()
# Select based on CV cutoff.
selCV <- function(object,cutoff) {
        cv.mean <- apply(object$cv,1,function(x) mean(x, na.rm=TRUE))
        return(na2false(cv.mean <= cutoff))
}
## na2false()
# Set all NA in a logical vector to FALSE.
na2false <- function(x) {
    x[which(is.na(x))] <- FALSE
    x	
}
## createWeights()
# Create weights based on Probe_type
# createWeights <- function(object, type.weights=NULL) {
# 	if(is.null(type))
# 		type.weights = c("DISCOVERY"=1,"FIDUCIAL"=0,"POSITIVE"=0,"NEGATIVE"=0, "OTHER"=0)
# 	
# 	tw = type.weights
# 	
# 	w <- array(1, dim(object))
# 	discovery <- object$type=="DISCOVERY"
# 	fiducial <- object$type=="FIDUCIAL"
# 	positive <- object$type=="POSITIVE"
# 	negative <- object$type=="NEGATIVE"
# 	other <- object$type=="OTHER"
# 	if(!is.null(type$DISCOVERY)) w[discovery,] <- tw$DISCOVERY
# 	if(!is.null(type$FIDUCIAL)) w[fiducial,] <- tw$FIDUCIAL
# 	if(!is.null(type$POSITIVE)) w[positive,] <- tw$POSITIVE
# 	if(!is.null(type$NEGATIVE)) w[negative,] <- tw$NEGATIVE
# 	if(!is.null(type$OTHER)) w[other,] <- tw$OTHER
# 	w
# }


setMethod("summary", "Codelink",
function(object, ...)
{
	summaryFlag(object)
})

setGeneric("summaryFlag", function(object) standardGeneric("summaryFlag"))
setMethod("summaryFlag", "Codelink",
function(object)
{
	flags <- c("G", "L", "S", "C", "I", "M", "X")
	res <- matrix(NA, length(flags), ncol(object),
		dimnames=list(flags, object$sample))
	for(flag in flags) {
		res[flag,] <- apply(object$flag, 2, function(z) length(grep(flag, z)))
	}
	res
})

# TODO: Add comment
# 
# Author: diez
###############################################################################


setClass("CodelinkSetUnique", contains="ExpressionSet")
setMethod("initialize", "CodelinkSetUnique",
function(.Object, sd = new("matrix"), ...)
{
	callNextMethod(.Object,	sd = sd, ...)
})

setValidity("CodelinkSetUnique", function(object)
{
	assayDataValidMembers(assayData(object), c("sd"))
})

TYPE_COLOR = c(
	DISCOVERY = "black",
	POSITIVE = "gray",
	NEGATIVE = "gray",
	OTHER = "gray",
	FIDUCIAL = "gray"
)

TYPE_BG = c(
	DISCOVERY = "black",
	POSITIVE = "blue",
	NEGATIVE = "red",
	OTHER = "green",
	FIDUCIAL = "yellow"
)

TYPE_PCH = list(
	DISCOVERY = ".",
	POSITIVE = 21,
	NEGATIVE = 21,
	OTHER = 21,
	FIDUCIAL = 21
)
# plotMA()
# MA plot of gene intensities.
plotMA <- function(object, array1 = 1, array2 = NULL, cutoff = c(-1, 1),
	label = NULL, type = NULL, high.list = NULL, high.col = "blue",
	high.pch = 21, high.bg = "cyan", snr = NULL, snr.cutoff = 1, 
	legend.x = NULL, pch = ".", subset = NULL, title = NULL, xlim = NULL,
	ylim = NULL)
{

	if(!is.null(type) && length(type) != dim(object)[1])
		warning("'type' has different length than object.")
	
	# get values.
	switch(class(object),
		Codelink = {
			islog <- object$method$log
			if(!is.null(object$Smean)) {
					val1 <- object$Smean[,array1]
					if(is.null(array2)) {
						val2 <- rowMeans(if(!islog) object$Smean
							else 2**object$Smean, na.rm = TRUE)
						if(islog) val2 <- log2(val2)
					} else
						val2 <- object$Smean[,array2]
			}
			if(!is.null(object$Ri)) {
					val1 <- object$Ri[,array1]
					if(is.null(array2)) {
						val2 <- rowMeans(if(!islog) object$Ri else
							2**object$Ri, na.rm = TRUE)
						if(islog) val2 <- log2(val2)
					} else
						val2 <- object$Ri[,array2]
			}
			if(!is.null(object$Ni)) {    
					val1 <- object$Ni[,array1]
					if(is.null(array2)) {
						val2 <- rowMeans(if(!islog) object$Ni else
							2**object$Ni, na.rm = TRUE)
						if(islog) val2 <- log2(val2)
					} else
						val2 <- object$Ni[,array2]
			}
			if(!islog) {
				val1 <- log2(val1)
				val2 <- log2(val2)
			}
			M <- val2 - val1
			A <- (val2 + val1)/2
			type <- object$type
			if(is.null(snr))
				snr <- rowMeans(object$snr, na.rm = TRUE)
			if(is.null(label) && !is.null(type))
				label = "type"
			if(is.null(array2))
				names <- paste("Mean Array vs.", object$sample[array1])
			else
				names <- paste(object$sample[array2], "vs.",
					object$sample[array1])
		},
		MArrayLM = {
			M <- as.matrix(object$coefficients)[, array1]
			A <- object$Amean
			# SNRmean can be added manually to the MArrayLM objects.
			if(is.null(snr))
				snr <- object$SNRmean
			names <- colnames(object$contrasts)[array1]
		},
		stop("invalid parameter object")
	)

	# check label.
	label <- match.arg(label,c("type", "snr", "none"))
	if(is.null(type) && label == "type") {
		warning("missing 'type' information for labelling spots.\n")
		label <- "none"
	}
	if(is.null(snr) && label == "snr") {
		warning("missing 'snr' information for labelling spots.\n")
		label <- "none"
	}

	# type based subset.
	if(!is.null(subset)) {
		if(is.null(type)) stop("need type information for subset.\n")
		else {
			subset <- match.arg(subset, levels(as.factor(type)), several.ok=TRUE)
			subset.sel <- type %in% as.character(subset)
			M <- M[subset.sel]
			A <- A[subset.sel]
			type <- type[subset.sel]
			snr <- snr[subset.sel]
			if(!is.null(high.list)) high.list <- high.list[subset.sel]
		}
	}

    # plot range.
	if(is.null(xlim)) xlim <- range(A, na.rm = TRUE)
	if(is.null(ylim)) ylim <- range(M, na.rm = TRUE)

	# basic plot.
	plot(0, col = "white", xlim = xlim, ylim = ylim, xlab="A", ylab="M")
	abline(h = 0, col = "steelblue", lwd = 2)
	if(!is.null(cutoff)) {
		abline(h=-cutoff, col = "gray", lty = "dotted")
		abline(h=cutoff, col = "gray", lty = "dotted")
	}

	# plot.
	switch(label,
		type = {
			for(level in names(TYPE_COLOR)) {
				sel <- type == level
				points(A[sel], M[sel], col = TYPE_COLOR[level],
					pch = TYPE_PCH[[level]], bg = TYPE_BG[level])
			}
			legend.text <- names(TYPE_COLOR)
			legend.fill <- TYPE_BG
		},
		snr = {
			g <- colorRampPalette(c("red", "orange"))
			s <- c(0, 0.85, snr.cutoff)
			col <- g(length(s) - 1)

			legend.text <- c()
			legend.fill <- c()

			for(n in 1:(length(s) - 1)) {
				sel <- snr >= s[n] & snr < s[n + 1]
				points(A[sel], M[sel], col = col[n], pch = pch);
				legend.text <- c(legend.text,
					paste(format(s[n], digits = 2, nsmall = 2), "<= SNR <",
					format(s[n + 1], digits = 2, nsmall = 2)))
				legend.fill <- c(legend.fill, col[n])
			}
			sel <- snr >= snr.cutoff
			points(A[sel], M[sel], col = "black", pch = pch);
			legend.text <- c(legend.text, paste("SNR >=", snr.cutoff))
			legend.fill <- c(legend.fill, "black")
		},
		none = {
			points(A, M, pch=pch);
		}
	)

	# lowess line block.
	# remove NA.
	sel <- which(!is.na(M))
	M.l <- M[sel]
	A.l <- A[sel]
	# take sample.
	subset = sample(1:length(M.l), min(c(5000, length(M.l))))
	A.l <- A.l[subset]
	M.l <- M.l[subset]
	# order it and remove duplicates.
	o <- order(A.l[subset])
	o <- which(!duplicated(A.l))
	# draw loess line.
	lines(approx(lowess(A.l[o], M.l[o])), col = "green", lwd=4)
	
	# highligh genes.
	if(!is.null(high.list))
		points(A[high.list], M[high.list], col=high.col, pch=high.pch,
			bg=high.bg)

	# title.
	if(is.null(title)) title(names)
	else title(title)

	# guess legend position.
	if(is.null(legend.x)) {
		if(abs(ylim[1]) > abs(ylim[2]))
			legend.x <- "bottomright"
		else
			legend.x <- "topright"
	}
		
	if(label != "none") legend(x = legend.x, legend = legend.text,
		fill = legend.fill, bty = "n")
}

# basic plotxy function (used by codPlotMA and codPlotScatter)
plotxy <- function(x, y, label = "type", cutoff = c(-1, 1), 
	snr.cutoff = 1, legend.x, legend.cex = 1, pch = ".", xlim, ylim, type, snr,
	high.list, title, loess = TRUE, xyline = FALSE, xlab = "x", ylab = "y", ...)
{

	# fix for data comming directly from Codelink suite.
	x <- x[is.finite(x)]
	y <- y[is.finite(y)]
	
	label <- match.arg(label, c("type", "snr", "none"))
	
	# basic plot.
	if(missing(xlim)) xlim <- range(x, na.rm = TRUE)
	if(missing(ylim)) ylim <- range(y, na.rm = TRUE)

	plot(0, col = "white", xlim = xlim, ylim = ylim, xlab = xlab, ylab = ylab)
	
	if(!is.null(cutoff)) {
		if(loess) {
			abline(h = 0, col = "grey", lwd = 2)
			abline(h = cutoff, col = "gray", lty = "dotted")
		}
	
		if (xyline) {
			abline(a = 0, b = 1, col = "gray", lwd = 2)
			abline(a = cutoff[1], b = 1, col = "gray", lty = "dotted")		
			abline(a = cutoff[2], b = 1, col = "gray", lty = "dotted")
		}
	}
	

	# plot XY values.
	switch(label,
		type = {
			#type <- probeTypes(x)
			for(level in names(TYPE_COLOR)) {
				sel <- type == level
				points(x[sel], y[sel], col = TYPE_COLOR[level],
					pch = TYPE_PCH[[level]], bg = TYPE_BG[level], ...)
			}
			legend.text = names(TYPE_COLOR)
			legend.fill = TYPE_BG
		},
		snr = {
			#snr <- meanSNR(x)
			s <- c(0, 0.85, snr.cutoff)
			col <- c("red", "orange")

			legend.text <- c()
			legend.fill <- c()

			for(n in 1:(length(s) - 1)) {
				sel <- snr >= s[n] & snr < s[n + 1]
				points(x[sel], y[sel], col = col[n], pch = pch, ...)
				legend.text <- c(legend.text,
					paste(format(s[n], digits = 2, nsmall = 2), "<= SNR <",
					format(s[n + 1], digits = 2, nsmall = 2)))
				legend.fill <- c(legend.fill, col[n])
			}

			sel <- snr >= snr.cutoff
			points(x[sel], y[sel], col = "black", pch = pch, ...)
			legend.text <- c(legend.text, paste("SNR >=", snr.cutoff))
			legend.fill <- c(legend.fill, "black")
		},
		none = points(x, y, pch = pch, ...)
	)
	
	if (loess) {
		# lowess line block.
		# remove NA.
		sel <- which(!is.na(y))
		y.l <- y[sel]
		x.l <- x[sel]
		# take sample.
		subset = sample(1:length(y.l), min(c(5000, length(y.l))))
		x.l <- x.l[subset]
		y.l <- y.l[subset]
		# order it and remove duplicates.
		o <- order(x.l[subset])
		o <- which(!duplicated(x.l))
		# draw loess line.
		lines(approx(lowess(x.l[o], y.l[o])), col = "green", lwd = 4)
	}
	
	# highligh genes.
	high.pch <- 21
	high.col <- "blue"
	high.bg <- "cyan"
	if(!missing(high.list))
		points(x[high.list], y[high.list], col=high.col, pch=high.pch, bg=high.bg)

	# title.
	if(!missing(title)) title(title)

	# guess legend position.
	if(missing(legend.x)) {
		if(abs(ylim[1]) > abs(ylim[2]))
			legend.x <- "bottomright"
		else
			legend.x <- "topright"
	}

	if(label != "none")
		legend(x = legend.x, legend = legend.text, fill = legend.fill, cex = legend.cex, bty = "n")
}

### plotMA()
## MA plot of gene intensities.
#plotMA <- function(object, array1=1, array2=2, cutoff=NULL, label="type",
	#type=NULL, high.list=NULL, high.col="gray", high.pch=21, high.bg="orange",
	#snr.cutoff=1, legend.x="bottomright", pch=".", subset=NULL, title=NULL, 
	#xlim=NULL, ylim=NULL)
#{
	##if(!is(object,"Codelink")) stop("Codelink object needed.")
##	if(!is.null(high.list) && (!is(high.list,"logical") || !is(high.list,"vector"))) stop("logical vector needed.")
##	if(!is.null(high.list) && length(high.list) != dim(object)[1]) stop("high.list and number of genes differ.")

	#switch(class(object),
		#Codelink={
			#if(!is.null(object$Smean)) {
					#val1 <- object$Smean[,array1]
					#val2 <- object$Smean[,array2]
					#what <- "Smean"
			#}
			#if(!is.null(object$Ri)) {
					#val1 <- object$Ri[,array1]
					#val2 <- object$Ri[,array2]
					#what <- "Ri"
			#}
			#if(!is.null(object$Ni)) {    
					#val1 <- object$Ni[,array1]
					#val2 <- object$Ni[,array2]
					#what <- "Ni"
			#}
			#if(!object$method$log) {
				#val1 <- log2(val1)
				#val2 <- log2(val2)
			#}
			## M, A computation.
			#M <- val2 - val1
			#A <- (val2 + val1)/2
			#type <- object$type
			#snr <- object$snr
		#},
		#MArrayLM={
			#M <- as.matrix(object$coefficients)[, array1]
			#A <- object$Amean
			#what <- "MArrayLM"
		#},
		#stop("invalid parameter object")
	#)
	
	#label <- match.arg(label,c("type", "snr", "none"))
	#if(is.null(type) && label!="none") {
		#cat("no type information, reverting to label 'none'\n")
		#label <- "none"
	#}

	#if(!is.null(subset)) {
		#if(is.null(type)) stop("need type information for subset.\n")
		#else {
			#subset <- match.arg(subset, levels(as.factor(type)), several.ok=TRUE)
			#subset.sel <- type %in% as.character(subset)
			#M <- M[subset.sel]
			#A <- A[subset.sel]
			#type <- type[subset.sel]
			#if(class(object)=="Codelink") snr <- snr[subset.sel,]
			#if(!is.null(high.list)) high.list <- high.list[subset.sel]
		#}
	#}


        ## Range computation.
	#if(is.null(xlim)) xlim <- range(A, na.rm=TRUE)
	#if(is.null(ylim)) ylim <- range(M, na.rm=TRUE)
	## Plotting.
	#switch(label,
                #type = {
                        #negative <- type=="NEGATIVE"
                        #positive <- type=="POSITIVE"
                        #discovery <- type=="DISCOVERY"
                        #fiducial <- type=="FIDUCIAL"
                        #other <- type=="OTHER"
                        #plot(A[discovery], M[discovery], xlim=xlim, ylim=ylim, xlab="A", ylab="M", pch=pch)
                        #points(A[negative], M[negative], col="red", pch=20)
                        #points(A[positive], M[positive], col="blue", pch=20)
                        #points(A[fiducial], M[fiducial], col="yellow", pch=20)
                        #points(A[other], M[other], col="green", pch=20)
			#legend.text <- c("DISCOVERY", "NEGATIVE", "POSITIVE", "FIDUCIAL", "OTHER")
			#legend.fill <- c("black","red","blue","yellow","green")
		#},
		#snr = {
			#sel.1 <- snr[, array1] >= snr.cutoff
			#sel.2 <- snr[, array2] >= snr.cutoff
			#plot(A[sel.1 & sel.2], M[sel.1 & sel.2], xlim=xlim, ylim=ylim, col="black", xlab="A", ylab="M", pch=pch)
                        #points(A[xor(sel.1, sel.2)], M[xor(sel.1, sel.2)], col="orange", pch=".")
                        #points(A[!sel.1 & !sel.2], M[!sel.1 & !sel.2], col="red", pch=".")
			#legend.text <- c("SNR >= 1 in all","SNR >= 1 in any","SNR < 1 in all")
			#legend.fill <- c("black", "orange", "red")
		#},
		#none = {
			#plot(A, M, xlab="A", ylab="M", pch=pch);
		#}
	#)

	## Misc.
    #abline(h=0, col="steelblue")
	#if(!is.null(cutoff)) {
	        #abline(h=-cutoff, col = "gray", lty="dotted")
        	#abline(h=cutoff, col = "gray", lty="dotted")
	#}
	## Highlighted genes.
	#if(!is.null(high.list)) {
		##names <- object$name[high.list]
		##text(A[high.list], M[high.list], names, col="blue", cex=0.75)
		#points(A[high.list], M[high.list], col=high.col, pch=high.pch,
			#bg=high.bg)
	#}

	### Lowess line.
	## Remove NA.
	#sel <- which(!is.na(M))
	#M <- M[sel]
	#A <- A[sel]
	## Take a sample.
	#subset=sample(1:length(M),min(c(10000, length(M))))
	#A <- A[subset]
	#M <- M[subset]
	## Order it and remove duplicates.
	#o <- order(A[subset])
	#o <- which(!duplicated(A))
	## draw the line.
	#lines(approx(lowess(A[o], M[o])), col = "green", lwd=4)
	
	#switch(class(object),
		#Codelink={
			#names <- paste(object$sample[array2],"-",object$sample[array1], sep="")
		#},
		#MArrayLM={
			#names <- colnames(object$contrasts)[array1]
		#}
	#)
	##if(is.null(title)) title(paste(names, " MA Plot of ", what, sep=""))
	#if(is.null(title)) title(names)
	#else title(title)
	#if(label != "none") legend(x=legend.x, legend=legend.text, fill=legend.fill, inset=0.05)
#}

## plotDensities()
# Densities plot of gene intensities.
# plotDensities <- function(object, subset=1:dim(object)[2], title=NULL, legend.cex=1, what=NULL) {
        # if(!is(object,"Codelink")) stop("Codelink object needed.")
	# if(is.null(what)) {
		# if(!is.null(object$Smean)) {
			# val <- object$Smean
			# what <- "Smean"
		# }
		# if(!is.null(object$Ri)) {
			# val <- object$Ri
			# what <- "Ri"
		# }
		# if(!is.null(object$Ni)) {
			# val <- object$Ni
			# what <- "Ni"
		# }
	# } else {
		# what <- match.arg(what, c("bg","snr","smean","ri","ni"))
		# switch(what,
			# snr={
				# val <- object$snr
				# what <- "SNR"
			# },
			# bg={
				# val <- object$Bmedian
				# what <- "Bmedian"
			# },
			# smean={
				# val <- object$Smean
				# what <- "Smean"
			# },
			# ri={
				# val <- object$Ri
				# what <- "Ri"
			# },
			# ni={
				# val <- object$Ni
				# what <- "Ni"
			# }
		# )
	# }
	# if(what!="Ni") val <- log2(val)
	# if(what=="Ni" && !object$method$log) val <- log2(val)
	# colors <- rainbow(length(subset))
	# y.max <- c()
        # x.max <- c()
        # for(n in subset) {
                # y.max[n] <- max(density(val[,n], na.rm=TRUE)$y)
                # x.max[n] <- max(density(val[,n], na.rm=TRUE)$x)
        # }
        # y.pos <- order(y.max, decreasing=TRUE, na.last=NA)
        # for(n in y.pos) {
                # k <- which(y.pos==n)
                # if(n==y.pos[1]) plot(density(val[,n], na.rm=TRUE), col=colors[k], main="")
                # else lines(density(val[,n], na.rm=TRUE), col=colors[k])
        # }
	# if(is.null(title)) title(paste("Density Plot of",what))
	# else title(title)
	# legend(x="topright", legend=object$sample[subset], cex=legend.cex, fill=colors, inset=0.05)
# }

plotDensities = function(object, what = NULL, title = NULL, col = NULL, legend.title = NULL, legend.cex=1, ...) 
{
if(!is(object,"Codelink")) stop("Codelink object needed.")

	# choose what to plot.
    if(is.null(what)) { # if what is null plot the most processed data available.
        if(!is.null(object$Smean)) {
            val <- object$Smean
            what <- "Smean"
        }
        if(!is.null(object$Ri)) {
            val <- object$Ri
            what <- "Ri"
        }
        if(!is.null(object$Ni)) {
            val <- object$Ni
            what <- "Ni"
        }
    } else { # or let the user choose)
        what <- match.arg(what, c("bg","snr","smean","ri","ni"))
        switch(what,
            snr={
                val <- object$snr
                what <- "SNR"
            },
            bg={
                val <- object$Bmedian
                what <- "Bmedian"
            },
            smean={
                val <- object$Smean
                what <- "Smean"
            },
            ri={
                val <- object$Ri
                what <- "Ri"
            },
            ni={
                val <- object$Ni
                what <- "Ni"
            }
        )
    }
    if(what!="Ni") val <- log2(val)
    if(what=="Ni" && !object$method$log) val <- log2(val)

    if(is.null(col))
        col <- rainbow(ncol(val))

	# compute densities (and limits).
	xlim = c()
    ylim = c()
    d = list()
    for(n in 1:ncol(val)) {
    	d[[n]] = density(val[,n], na.rm=TRUE)
    	xlim = c(xlim, d[[n]]$x)
    	ylim = c(ylim, d[[n]]$y)
   	}
   	xlim = range(xlim, na.rm = TRUE)
   	ylim = range(ylim, na.rm = TRUE)
   	
   	# plot densities.
    for(n in 1:ncol(val)) {
        if (n==1)
        	plot(d[[n]], col=col[n], main="", xlim = xlim, ylim = ylim, ...)
        else
        	lines(d[[n]], col=col[n], ...)
	}
    if(is.null(title)) title(paste("Density Plot of",what))
    else title(title)
    legend(x="topright", legend=object$sample, cex=legend.cex, fill=col, title = legend.title)	
}



## plotCorrelation()
# Scatter plot of intensities: One array in x compare to other in y axis.
plotCorrelation <- function(object, x=1, y=2, cutoff=FALSE, label="type", title=NULL, xlim=NULL, ylim=NULL) {
        if(!is(object,"Codelink")) stop("Codelink object needed.")       
        if(!is.null(object$Smean)) {
                        xval <- object$Smean[,x]
                        yval <- object$Smean[,y]
			what <- "Smean"
        }       
        if(!is.null(object$Ri)) {
                        xval <- object$Ri[,x]
                        yval <- object$Ri[,y]
			what <- "Ri"
        } 
        if(!is.null(object$Ni)) {    
                        xval <- object$Ni[,x]
                        yval <- object$Ni[,y]
			what <- "Ni"
        }
	if(!object$method$log) {
		xval <- log2(xval)
		yval <- log2(yval)
	}
        names <- object$sample[c(x,y)]
	 # Range computation.
        if(is.null(xlim)) xlim <- range(xval, na.rm=TRUE)
        if(is.null(ylim)) ylim <- range(yval, na.rm=TRUE)
        
	switch(label,
                type={
                        negative <- object$type=="NEGATIVE"
                        positive <- object$type=="POSITIVE"
                        discovery <- object$type=="DISCOVERY"
                        fiducial <- object$type=="FIDUCIAL"
                        other <- object$type=="OTHER"
                        plot(xval[discovery], yval[discovery], xlim=xlim, ylim=ylim, xlab=names[1], ylab=names[2], pch=".")
                        points(xval[negative],yval[negative],col="red",pch=20)
                        points(xval[positive],yval[positive],col="blue",pch=20)
                        points(xval[fiducial],yval[fiducial],col="yellow",pch=20)
                        points(xval[other],yval[other],col="green",pch=20)
			legend.text <- c("DISCOVERY","NEGATIVE","POSITIVE","FIDUCIAL","OTHER")
			legend.fill <- c("black","red","blue","yellow","green")
                },
		snr = {
                        foo.mean <- apply(object$snr[, c(x,y)], 1, mean)
                        foo.sel <- foo.mean >= 1
                        plot(xval[foo.sel], yval[foo.sel], xlim=xlim, ylim=ylim, col="black", xlab=names[1], ylab=names[2], pch=".")
                        points(xval[!foo.sel],yval[!foo.sel],col="orange",pch=".")
			legend.text <- c("SNR >= 1","SNR < 1")
			legend.fill <- c("black","orange")
                },
                none = {
			plot(xval, yval, pch=".", xlab=names[1], ylab=names[2])
			#legend.text <- "ALL PROBES"
			#legend.fill <- "black"
                }
        )

	abline(0,1,col="blue")
	if(!is.null(cutoff)) {
		abline(1,1, lty="dotted")
		abline(-1,1, lty="dotted")
	}
	#if(is.null(title)) title(paste(object$Experiment_name,"\n",names[1]," vs ", names[2], " - Scatter Plot of",what))
	if(is.null(title)) title(paste(names[1]," vs ", names[2], " - Scatter Plot of",what))
	else title(title)
	if(label != "none") legend(x="topleft", legend=legend.text, fill=legend.fill, inset=0.05)
}

## plotCV()
# density plot of C.V. of merged objects.
plotCV <- function(object, subset=c(1:dim(object)[2]), cutoff=NULL, title=NULL, legend.cex=1) {
	if(!is(object,"Codelink")) stop("Codelink object needed.")
	if(object$method$merge == "NONE") stop("Merged object needed.")
	colors <- rainbow(length(subset))
	y.max <- c()
	x.max <- c()
	for(n in subset) {
        	y.max[n] <- max(density(object$cv[,n],na.rm=TRUE)$y)
	        x.max[n] <- max(density(object$cv[,n],na.rm=TRUE)$x)
	}
	y.pos <- order(y.max,decreasing=TRUE,na.last=NA)
	for(n in y.pos) {
        	k <- which(y.pos==n)
	        if(n==y.pos[1]) plot(density(object$cv[,n],na.rm=TRUE),col=colors[k], main="")
        	else lines(density(object$cv[,n],na.rm=TRUE),col=colors[k])
	}
	if(is.null(title)) title("CV of Merged Ni")
	else title(title)
	legend(x="topright", legend=object$sample[subset], cex=legend.cex, fill=colors, inset=0.05)
	if(!is.null(cutoff)) abline(v=cutoff, lty="dotted")
}

# Generate red-black-green colorscale
#redgreen <- colorRampPalette(c('red','black','green'))
# Generate green-black-red colorscale
#greenred <- colorRampPalette(c('green','black','red'))
# Generate blue white red colorscale
#bluered <- colorRampPalette(c('blue','white','red'))

## imageCodelink()
# Function to plot images of arrays.
imageCodelink <- function (object, array = 1, what = "bg",
    low="black", high="white", mar = c(1, 1, 1, 1),
    gr=1, gc=1, log.it=FALSE, ...) {

	what <- match.arg(what, c("bg", "snr", "smean", "ri", "ni"))
	switch(what,
		bg = val <- object$Bmedian[, array],
		snr = val <- object$snr[, array],
		smean = val <- object$Smean[, array],
		ri = val <- object$Ri[,	array],
		ni = val <- object$Ni[, array]
	)
	if(!object$method$log && log.it)
		val <- log2(val)

	o.row <- range(object$logical[, "row"])
	o.col <- range(object$logical[, "col"])
	
	foo <- matrix(NA, nrow = o.row[2], ncol = o.col[2])

	for (n in 1:dim(object)[1]) {
		foo[object$logical[n, "row"],object$logical[n, "col"]] <- val[n]
	}

	col <- colorRampPalette(c(low, high))(123)
	old.par <- par(mar = mar)
	on.exit(par(old.par))

	sc <- o.col[2]/gc
	sr <- o.row[2]/gr

	image(0:(gr * sr), 0:(gc * sc), foo, col = col, xaxt="n", yaxt="n", ...)

	for (igrid in 0:gc) lines(c(0, gr * sr), rep(igrid * sc, 2))
		for (igrid in 0:gr) lines(rep(igrid * sr, 2), c(0, gc * sc))
	mtext(paste("Slide:", array," File:", object$file[array]), side=1, cex=0.8)
}

# arrayNew
# creates a suitable x11 device to see the chip with the correct dimensions.
arrayNew <- function(f = 2, chip = "rwgcod") {
	d <- getChipDimensions(chip)
	
	if(is.null(d))
		stop("unknown chip.")
	
	gc <- d["gc"]
	gr <- d["gr"]
	sc <- d["sc"]
	sr <- d["sr"]

	r <- sr / sc
	x11(width = f * gr * r, height = f * gc)
}

# genefilter function for filterSNR().
filtersnr <- function(p = 1, A = 1, na.rm = TRUE) {
	function(x) {
		if(na.rm)
			x <- x[!is.na(x)]
		sum(x < A)/length(x) >= p
	}
}

# filter probes when all samples have SNR < cutoff [default: 1]
setGeneric("filterSNR", function(x, cutoff = 1) standardGeneric("filterSNR"))
setMethod("filterSNR", "Codelink",
function(x, cutoff = 1)
{
	filterSNR(x$snr, cutoff = cutoff)
})
setMethod("filterSNR", "CodelinkSet",
function(x, cutoff = 1)
{
	filterSNR(getSNR(x), cutoff = cutoff)
})
setMethod("filterSNR", "matrix",
function(x, cutoff = 1)
{
	requireNamespace("genefilter")
	f <- filtersnr(A = cutoff)
	!genefilter::genefilter(x, genefilter::filterfun(f))
})

# codUniverse()
# create a universe of EG terms, applying a non-specific filter.
setGeneric("codUniverse", function(x, chip) standardGeneric("codUniverse"))
setMethod("codUniverse", "Codelink",
function(x, chip)
{
	if(missing(chip))
		stop("chip type missing.")
	# filter not expressed probes.
	filtersnr <- filterSNR(x)
	x <- x[filtersnr,]

	probesname <- x$name
	mediansnr <- apply(x$snr, 1, function(z) median(z, na.rm = TRUE))
    mediansnr <- data.frame(probeName = probesname, medianSNR = mediansnr)

	createUniverse(chip, probesname, mediansnr)
})
setMethod("codUniverse", "CodelinkSet",
function(x, chip)
{
	if(missing(chip))
		chip <- annotation(x)
	if(nchar(chip) < 1)
		stop("chip type missing.")

	# filter not expressed probes.
	filtersnr <- filterSNR(x)
	x <- x[filtersnr,]

	probesname <- probeNames(x)
	# in CodelinkSet is currently stored meanSNR, may be good to have both.
	mediansnr <- apply(getSNR(x), 1, function(z) median(z, na.rm = TRUE))
    mediansnr <- data.frame(probeName = probesname, medianSNR = mediansnr)

	createUniverse(chip, probesname, mediansnr)
})

# more or less general function.
createUniverse <- function(chip, probesname, mediansnr) {
	if(!do.call(require, list(chip)))
		stop("package", chip, "not found.")

	# probes with EG.
	envirEG <- get(paste(chip, "ENTREZID", sep = ""))
	eg <- mget(probesname, envir = envirEG)

	haveEG <- sapply(eg, function(x) !is.na(x))
	haveEG <- unique(names(which(haveEG)))

	# probes with GO.
	envirGO <- get(paste(chip, "GO", sep = ""))
	go <- mget(probesname, envir = envirGO)
	haveGO <- sapply(go, function(x)
		ifelse(length(x) == 1 && is.na(x), FALSE, TRUE))
	haveGO <- unique(names(which(haveGO)))

	allProbes <- intersect(haveEG, haveGO)

	egSubset <- eg[allProbes]
	egSubsetUnique <- unique(unlist(egSubset))
	allProbesUnique <- sapply(egSubsetUnique, function(x) {
		selEG <- egSubset == x
		probesName <- names(egSubset)[selEG]
		if(length(probesName) == 1) probesName
		else {
			tmp <- mediansnr[mediansnr[, "probeName"] %in% probesName, ]
			if(all(is.na(tmp[, 2]))) probesName[1]
			else {
				sel <- tmp[ ,"medianSNR"] == max(tmp[, "medianSNR"], na.rm = TRUE)
				sel <- na2false(sel)
				as.character(tmp[sel, "probeName"])
			}
		}
	})

	# now, get the ids from universe and check it's ok.
	egUniverse <- unlist(mget(allProbesUnique, envir = envirEG))
	if(any(duplicated(egUniverse)))
		stop("error in gene universe: can't have duplicated ids.")
	else
		egUniverse
}

# Codelink-class definition.
# Based on the definitions found in limma v1.8.20 for RGList objects.
setClass("Codelink", representation("list"))

# Addapted? from limma:
#  Print leading 5 elements or rows of atomic object.
#  Gordon Smyth
#  May 2003.  Last modified 7 April 2004.
printHead <- function(x) {
	if(is.atomic(x)) {
		d <- dim(x)
		if(length(d)<2) which <- "OneD"
		if(length(d)==2) which <- "TwoD"
		if(length(d)>2) which <- "Array"
	} else {
		if(inherits(x,"data.frame")) {
			d <- dim(x)
			which <- "TwoD"
		} else
		which <- "Recursive"
	}
	switch(which,
	OneD={
		n <- length(x)
		if(n > 20) {
			print(x[1:5])
			cat(n-5,"more elements ...\n")
		} else
		print(x)
	},
	TwoD={
		n <- d[1]
		if(n > 10) {
			print(x[1:5,])
			cat(n-5,"more rows ...\n")
		} else
		print(x)
	},
	Array={
		n <- d[1]
		if(n > 10) {
			dn <- dimnames(x)
			dim(x) <- c(d[1],prod(d[-1]))
			x <- x[1:5,]
			dim(x) <- c(5,d[-1])
			if(!is.null(dn[[1]])) dn[[1]] <- dn[[1]][1:5]
				dimnames(x) <- dn
				print(x)
				cat(n-5,"more rows ...\n")
		} else
		print(x)
	},
	Recursive=print(x)
	)
}

# show method.
setMethod("show","Codelink",
#  Print and show method large data objects
#  Based on same from: limma
#  Gordon Smyth
#  May 2003
function(object) {
	cat("An object of class \"",class(object),"\"\n",sep="")
	for (what in names(object)) {
		x <- object[[what]]
		cat("$",what,"\n",sep="")
		printHead(x)
		cat("\n")
	}
	for (what in setdiff(slotNames(object),".Data")) {
		x <- slot(object,what)
		if(length(x) > 0) {
			cat("@",what,"\n",sep="")
			printHead(x)
			cat("\n")
		}
	}
})

# [ method.
setMethod("[", "Codelink",
function(x, i, j, ..., drop=FALSE) {
	if(!missing(i)) {
		x$Smean <- x$Smean[i,, drop = FALSE]
		x$Bmedian <- x$Bmedian[i,, drop = FALSE]
		x$Bstdev <- x$Bstdev[i,, drop = FALSE]
		x$snr <- x$snr[i,,drop = FALSE]
		x$Ri <- x$Ri[i,,drop = FALSE]
		x$Ni <- x$Ni[i,,drop = FALSE]
		x$cv <- x$cv[i,,drop = FALSE]
		x$flag <- x$flag[i,,drop = FALSE]
		x$name <- x$name[i]
		x$type <- x$type[i]
		x$logical <- x$logical[i,,drop = FALSE]
	}
	if(!missing(j)) {
		x$Smean <- x$Smean[,j, drop = FALSE]
		x$Bmedian <- x$Bmedian[,j, drop = FALSE]
		x$Bstdev <- x$Bstdev[,j, drop = FALSE]
		x$snr <- x$snr[,j, drop = FALSE]
		x$Ri <- x$Ri[,j, drop = FALSE]
		x$Ni <- x$Ni[,j, drop = FALSE]
		x$cv <- x$cv[,j, drop = FALSE]
		x$flag <- x$flag[,j, drop = FALSE]
		x$file <- x$file[j]
		x$sample <- x$sample[j]
        }
        return(x)
})

# S3 methods.
# dim.
dim.Codelink <- function(x) {
	if(is.null(x$Ni) && is.null(x$Ri) && is.null(x$Smean)) return(c(0,0))
	if(!is.null(x$Smean)) return(dim(x$Smean))
	if(!is.null(x$Ri)) return(dim(x$Ri))
	if(!is.null(x$Ni)) return(dim(x$Ni))
}
# as.matrix.
as.matrix.Codelink <- function(x, ...) {
	if(is.null(x$Ni) && is.null(x$Ri) && is.null(x$Smean)) return(NULL)
	if(!is.null(x$Smean)) return(x$Smean)
	if(!is.null(x$Ri)) return(x$Ri)
	if(!is.null(x$Ni)) return(x$Ni)
}

# temporary wrapper.
#readCodelink2 <- function(..., phenodata = NULL, featuredata = NULL) {
#	#colnames = list(Signal = "Spot_mean", Background = "Bkgd_median")) {
#	tmp <- readCodelink(...)
#	#Codelink2eSet(tmp)
#	c2e(tmp, phenodata = phenodata, featuredata = featuredata)
#}

# readCodelinkSet <- function(targets, filename, columns = list(Signal = "Spot_mean", Background = "Bkgd_median"), phenoData, ...) {
# 	if(missing(filename) && missing(targets)) stop("argument 'targets' or 'filename' must be specified.")
#     if (missing(targets)) {
#         filename <- as.character(filename)
# 		if(missing(phenoData)) pdata <- NULL
# 		else pdata <- phenoData
#     } else {
#         pdata <- read.AnnotatedDataFrame(targets)
# 		if(! any(grep("FileName", varLabels(pdata)))) {
#             filename <- sampleNames(pdata)
#         } else {
#             filename <- pData(pdata)[, "FileName"]
#         }
#     }
#     if (is.null(filename)) stop("invalid filenames.")
#
# 	tmp <- readCodelink(files = filename, ...)
#
# 	c2e(tmp, phenodata = pdata, featuredata = NULL)
# }


readCodelinkSet <- function(filename, path, phenoData=NULL, ...) {
	if(missing(filename)) stop("argument 'filename' must be specified.")
	if (!missing(path)) filename <- file.path(path, filename)
	tmp = .readCodelinkRaw(files = filename, ...)
	Codelink2CodelinkSet(tmp, phenodata = phenoData, ...)
}


Codelink2CodelinkSet <- function (object, annotation = NULL, phenodata = NULL, featuredata = NULL, type = "Spot")
{
    if (class(object) != "Codelink")
        stop("Codelink-object needed.")

	if (type == "Spot" && is.null(object$Smean))
		stop("Smean is missing. Choose type Raw or Norm")
	if (type == "Raw" && is.null(object$Ri))
		stop("Ri is missing. Choose type Spot or Norm")
	if (type == "Norm" && is.null(object$Ni))
		stop("Ni is missing. Choose intensity Spot or Raw")
    switch(type,
    	"Spot" = int <- object$Smean,
		"Raw" = int <- object$Ri,
    	"Norm" = int <- object$Ni
    )
	if(is.null(object$Bmedian)) {
		bkg <- matrix(NA, ncol = ncol(int), nrow = nrow(int))
	}
	else
		bkg <- object$Bmedian

    if (is.null(phenodata)) {
	    phenodata <- data.frame(sample = object$sample)
    	phenodata.varMet <- data.frame(labelDescription = "sample names", row.names = "sample")
    	phenodata <- new("AnnotatedDataFrame", data = phenodata, varMetadata = phenodata.varMet)
    }

    if (is.null(featuredata)) {
    	featuredata <- data.frame(probeName = object$name, probeType = object$type, logicalRow = object$logical[, "row"], logicalCol = object$logical[, "col"], meanSNR = rowMeans(object$snr, na.rm = TRUE))
    	featuredata.feMet <- data.frame(labelDescription = c("probe names", "probe types", "probe row position", "probe column position", "mean snr"), row.names = c("probeName", "probeType", "logicalRow",      "logicalCol", "meanSNR"))
		featuredata <- new("AnnotatedDataFrame", data = featuredata, varMetadata = featuredata.feMet)

    }

    if (is.null(annotation))
    	chip <- annotation(object)

    if(is.null(object$weight))
    	object$weight=createWeights(object)

	tmp <- new("CodelinkSet", exprs = int, background = bkg,
        flag = object$flag, weight=object$weight, snr = object$snr, annotation = chip)
    phenoData(tmp) <- phenodata
    featureData(tmp) <- featuredata
	experimentData(tmp)@preprocessing <- object$method
	experimentData(tmp)@other <- list("product" = object$product)
    tmp
}

# convert and old Codelink object to an CodelinkRawSet object.
Codelink2eSet <- function(object, annotation = NULL) {
	if(class(object) != "Codelink") stop("Codelink-object needed.")

	pD <- data.frame(sample=unique(object$sample))
	varMet <- data.frame(labelDescription="sample names", row.names="sample")
	pD <- new("AnnotatedDataFrame", data=pD, varMetadata=varMet)

	Rep <- data.frame(probeName = object$name, probeType = object$type,
		logicalRow = object$logical[, "row"],
		logicalCol = object$logical[, "col"], meanSNR = rowMeans(object$snr, na.rm = TRUE))
	feMet <- data.frame(labelDescription = c("probe names", "probe types", "probe row position", "probe column position", "mean snr"),
		row.names = c("probeName", "probeType", "logicalRow", "logicalCol", "meanSNR"))
	Rep <- new("AnnotatedDataFrame", data = Rep, varMetadata = feMet)

	if(is.null(annotation))
		chip <- annotation(object)

	int <- NULL
	bkg <- NULL
	if(!is.null(object$Smean)) int <- object$Smean
	if(!is.null(object$Ri)) int <- object$Ri
	if(!is.null(object$Ni)) int <- object$Ni
	if(!is.null(object$Bmedian)) bkg <- object$Bmedian
	else bkg <- int

	#tmp <- new("CodelinkRawSet", phenoData = pD, intensity = int,
	#tmp <- new("CodelinkSet", phenoData = pD, intensity = int,
	tmp <- new("CodelinkSet", phenoData = pD,  exprs = int,
		background = bkg, flag = object$flag, snr = object$snr,
		annotation = chip, featureData = Rep)

	featureNames(tmp) <- object$id
	sampleNames(tmp) <- object$sample

	experimentData(tmp)@preprocessing <- object$method
	experimentData(tmp)@other <- list("product" = object$product)

	tmp
}

# TODO: this function makes the opposite conversion.
eSet2Codelink <- function(object, annotation=NULL)
{
	tmp <- new("Codelink")

	warning("FIXME: not implemented.")

	tmp
}

# guess the annotation package associated.
setMethod("annotation", "Codelink",
function(object)
{
	base <- "cod"
	tmp <- object$product

	# organism.
	org <- ""
	if(any(grep("Rat", tmp))) org <- "r"
	if(any(grep("Mouse", tmp))) org <- "m"
	if(any(grep("Human", tmp))) org <- "h"

	# chip.
	chip <- ""
	if(any(grep("Whole Genome", tmp))) chip <- "wg"
	if(any(grep("UniSet.*I", tmp))) chip <- "10k"
	if(any(grep("UniSet.*II", tmp))) chip <- "20k"

	# return name.
	if(org != "" && chip != "")
		ann <- paste(org, chip, base, ".db", sep="")
	else ann <- ""

	ann
})

# CodelinkSet-class.
setClass("CodelinkSet", contains="ExpressionSet")

setValidity("CodelinkSet", function(object) {
	assayDataValidMembers(assayData(object), c("background", "flag", "weight", "snr"))
})

## ----include=FALSE,cache=FALSE------------------------------------------------
library(codelink)
library(knitr)
opts_chunk$set(fig.align = 'center', concordance=TRUE,width=50,fig.show="hold", tidy = TRUE, comment = "", highlight = FALSE, prompt = TRUE)
knit_hooks$set(no.mar = function(before, options, envir) {
    if (before) par(mar = rep(0,4))  # no margins.
})

## ----eval=FALSE---------------------------------------------------------------
# # NOT RUN #
# library(codelink)
# # to read data as CodelinkSet object:
# f = list.files(pattern="TXT")
# codset = readCodelinkSet(filename=f)
# # NOT RUN #

## ----eval=FALSE---------------------------------------------------------------
# # NOT RUN #
# pdata=read.AnnotatedDataFrame("targets.txt")
# codset=readCodelinkSet(filename=pdata$FileName, phenoData=pdata)
# # NOT RUN #

## -----------------------------------------------------------------------------
# sample dataset.
data(codset)
codset

## -----------------------------------------------------------------------------
data(codelink.example)
print(is(codelink.example))
tmp=Codelink2CodelinkSet(codelink.example)
tmp

## -----------------------------------------------------------------------------
w = createWeights(codset)
## NOTE: a proper replacement function will be provided later:
assayDataElement(codset,"weight")=w

## -----------------------------------------------------------------------------
# get signal intensities. alias: getInt()
head(exprs(codset))

# get background intensities.
head(getBkg(codset))

# get SNR values.
head(getSNR(codset))

# get flags.
head(getFlag(codset))

# get weights.
head(getWeight(codset))

# get phenoData:
head(pData(codset))


## -----------------------------------------------------------------------------
codset = codCorrect(codset, method = "half", offset = 0)

## -----------------------------------------------------------------------------
codset = codNormalize(codset, method = "quantile")

## ----eval=FALSE---------------------------------------------------------------
# # NOT RUN
# codset = codNormalize(codset, method = "loess", weights=getWeight(codset), loess.method="fast")
# # NOT RUN

## ----fig.cap="MA plot (left) and density plot (right).",fig.height=4,fig.width=4,out.width='.49\\linewidth'----
codPlot(codset) # by default MA plot.
codPlot(codset, what="density")

## ----fig.cap="Pseudo image plot of an array",fig.height=1,fig.width=3---------
codPlot(codset, what="image")

## -----------------------------------------------------------------------------
fit = lmFit(codset, design=c(1,1,2,2), weights=getWeight(codset))
fit2 = eBayes(fit)
topTable(fit2)

## -----------------------------------------------------------------------------
citation(package="codelink")

## -----------------------------------------------------------------------------
sessionInfo()

# codCorrect-method.
setGeneric("codCorrect", function(object, method = "half", normexp.method="saddle", offset=0)	standardGeneric("codCorrect"))
setMethod("codCorrect", "CodelinkSet",
function(object, method = "half", normexp.method="saddle", offset=0)
{
	x = object
	if(getInfo(x, "background") != "NONE") {
		warning("data already corrected, skipped.")
		return(x)
	}
	
	method <- match.arg(method, c("none", "subtract", "half", "normexp"))

	# take data.
	#int <- assayDataElement(x, "intensity")
	#bkgd <- assayDataElement(x, "background")
	int <- getInt(x)
	bkgd <- getBkg(x)

	# correct intensities.
	newInt=backgroundCorrect.matrix(int, bkgd, method=method, normexp.method=normexp.method,offset=offset)

	assayDataElement(x, "exprs") <- newInt
	
	# info.
	experimentData(x)@preprocessing[["background"]] <- method
	
	x
})

# getInfo-method.
setGeneric("getInfo", function(x, what)
	standardGeneric("getInfo"))
setMethod("getInfo", "CodelinkSet",
function(x, what)
{
	if(missing(what))
		experimentData(x)@preprocessing
	else
		experimentData(x)@preprocessing[[what]]
})

# codNormalize-method.
setGeneric("codNormalize", function(object, method = "quantile", log.it = TRUE, ...)	standardGeneric("codNormalize"))
setMethod("codNormalize", "CodelinkSet",
function(object, method="quantile", log.it=TRUE, weights=NULL, loess.method="fast")
{	
	normalize(object,method=method,log.it=log.it,weights=weights,loess.method=loess.method)
})

setMethod("normalize", "CodelinkSet",
function(object, method="quantile", log.it=TRUE, weights=NULL, loess.method="fast")
{
	x = object
	method <- match.arg(method, c("median", "quantile", "loess"))

	if(getInfo(x, "normalization") != "NONE") {
		warning("data already normalized, skipped.")
		return(x)
	}
	
	# take data.
	#newInt <- assayDataElement(x, "intensity")
	newInt <- getInt(x)

	# log.it?
	if(log.it && !getInfo(x, "log"))
		newInt <- log2(newInt)
	else
		if(log.it && getInfo(x, "log"))
			warning("already log2 data, skipping...")

	# check weights.
	if(!is.null(weights))
		if(is.matrix(weights))
			weights=apply(weights,1,min)
	
	# do normalization stuff.
	switch(method,
		#median = newInt <- normalize.median(newInt,weights=weights),
		median = newInt <- normalizeMedianValues(newInt),
		quantile = newInt <- normalizeQuantiles(newInt),
		#quantile = newInt <- normalize.quantiles(newInt),
		#loess = newInt <- normalize.loess(newInt, log.it = FALSE,
		#	verbose = FALSE)
		loess = newInt <- normalizeCyclicLoess(newInt, weights=weights, method = loess.method)	 
	)

	# reassign data.
	assayDataElement(x, "exprs") <- newInt
	
	# info.
	experimentData(x)@preprocessing[["normalization"]] <- method
	experimentData(x)@preprocessing[["log"]] <- log.it

	x
})

# codPreprocess-method.
setGeneric("codPreprocess", function(x, bg.method = "half", 
	norm.method = "quantile", log.it = TRUE) standardGeneric("codPreprocess"))
setMethod("codPreprocess", "CodelinkSet",
function(x, bg.method = "half", norm.method = "quantile", log.it = TRUE)
{
	x <- codCorrect(x, method = bg.method)
	x <- codNormalize(x, method = norm.method, log.it = log.it)
	x
})

# exprs method.
setMethod("exprs", "CodelinkSet",
function(object)
{
	#getInt(object)
	assayDataElement(object, "exprs")
})

# getInt method.
setGeneric("getInt", function(object) standardGeneric("getInt"))
setMethod("getInt", "CodelinkSet",
function(object)
{
	exprs(object)
	#assayDataElement(object, "intensity")
})

# getBkg method.
setGeneric("getBkg", function(object) standardGeneric("getBkg"))
setMethod("getBkg", "CodelinkSet",
function(object)
{
	assayDataElement(object, "background")
})

# getSNR method.
setGeneric("getSNR", function(object) standardGeneric("getSNR"))
setMethod("getSNR", "CodelinkSet",
function(object)
{
	assayDataElement(object, "snr")
})

# getFlag method.
setGeneric("getFlag", function(object) standardGeneric("getFlag"))
setMethod("getFlag", "CodelinkSet",
function(object)
{
	assayDataElement(object, "flag")
})

# getWeight method.
#setMethod("weights", "CodelinkSet",
#function(object)
# {
#	assayDataElement(object, "weight")
#})
# alias.
setGeneric("getWeight", function(object) standardGeneric("getWeight"))
setMethod("getWeight", "CodelinkSet",
function(object)
{
	#weights(object)
	assayDataElement(object, "weight")
})

# assignWeight method.
#setGeneric("weights<-", function(object, value) standardGeneric("weights<-"))
#setReplaceMethod("weights", signature(object="CodelinkSet",value="matrix"),
#function(object, value) assayDataElementReplace(object, "weight", value))

# probeNames-method.
setGeneric("probeNames", function(object) standardGeneric("probeNames"))
setMethod("probeNames", "CodelinkSet",
function(object)
{
	as.character(pData(featureData(object))[, "probeName"])
})

# probeTypes method to get feature types.
setGeneric("probeTypes", function(object) standardGeneric("probeTypes"))
setMethod("probeTypes", "CodelinkSet",
function(object)
{
	as.character(pData(featureData(object))[, "probeType"])
})

# meanSNR method to get mean SNR values.
setGeneric("meanSNR", function(object) standardGeneric("meanSNR"))
setMethod("meanSNR", "CodelinkSet",
function(object)
{
	pData(featureData(object))[, "meanSNR"]
})

# logicalRow method to get chip row positions.
setGeneric("logicalRow", function(object) standardGeneric("logicalRow"))
setMethod("logicalRow", "CodelinkSet",
function(object)
{
	pData(featureData(object))[, "logicalRow"]
})

# logicalCol method to get chip col positions.
setGeneric("logicalCol", function(object) standardGeneric("logicalCol"))
setMethod("logicalCol", "CodelinkSet",
function(object)
{
	pData(featureData(object))[, "logicalCol"]
})

# experimental codPlot-method.
# this intends to be a general interface for all plotting utilities.
setGeneric("codPlot", function(object, array, what = "ma", ...)
	standardGeneric("codPlot"))

# codPlot, CodelinkSet-method.
setMethod("codPlot", "CodelinkSet",
function(object, array, what = "ma", ...)
{
	x = object
	what <- match.arg(what, c("ma", "density", "scatter", "image"))

	switch(what,
		ma = {
			if(missing(array)) array = 1;
			codPlotMA(x, array, ...) },
		density = {
			#if(missing(array)) array = NULL
			#codPlotDensity(x, array, ...) },
			codPlotDensity(x, array=array, ...) },
		scatter = {
			if(missing(array)) array = 1;
			codPlotScatter(x, array=array, ...) },
		image = {
			if(missing(array)) array = 1;
			codPlotImage(x, array=array, ...) }
	)
})
# codPlot(), MArrayLM-method.
setMethod("codPlot", "MArrayLM",
function(object, array, what = "ma", ...)
{
	if(missing(array)) array = 1
	codPlotMA(object, array = array, ...)
})

# codPlotMA, generic.
setGeneric("codPlotMA", function(x, array = 1, ...)
	standardGeneric("codPlotMA"))
# codPlotMA, CodelinkSet-method.
setMethod("codPlotMA", "CodelinkSet",
function(x, array = 1, array2, label = "type", cutoff = c(-1, 1),
	snr.cutoff = 1,	legend.x, pch = ".", ...)
{
	label <- match.arg(label, c("type", "snr", "none"))
	type <- probeTypes(x)
	snr <- meanSNR(x)

	# compute MA values.
	X <- getInt(x)
	islog <- getInfo(x, "log")
	
	if (!islog) X <- log2(X)
	
	X1 <- X[, array]
	if (missing(array2))
		X2 <- rowMeans(X[, -array, drop = FALSE], na.rm = TRUE)
	else
		X2 <- X[, array2]

	M <- X2 - X1
	A <- (X1 + X2) / 2

	# get title.
	samples <- sampleNames(x)
	if (missing(array2)) {
		title <- paste("Median vs", samples[array])
	} else {
		title <- paste(samples[array2], "vs", samples[array])
	}

	plotxy(A, M, label = label,	cutoff = cutoff, snr.cutoff = snr.cutoff,
		legend.x = legend.x, pch = pch, type = type, snr = snr, title = title, 
		xlab = "A", ylab = "M", ...)
})
# codPlotMA, MArrayLM-method.
setMethod("codPlotMA", "MArrayLM",
function(x, array = 1, label = "type", cutoff = c(-1, 1),
	snr.cutoff = 1, legend.x, pch = ".", Amean, type, snr, ...)
{
	label <- match.arg(label, c("type", "snr", "none"))
	
	if(missing(type) && label == "type") {
		label = "none"
		warning("missing type information.")
	}   
	if(missing(snr) && label == "snr") {
		label = "none" 
		warning("missing snr info.")
	}

	M <- x$coef[, array]
	if(is.null(x$Amean))
		if(missing(Amean)) stop("no Amean information.")
		else A <- Amean
	else
		A <- x$Amean
	
	title <- colnames(x$coef)[array]
	
	plotxy(A, M, label = label,	cutoff = cutoff, snr.cutoff = snr.cutoff,
		legend.x = legend.x, pch = pch, type = type, snr = snr, title = title, 
		xlab = "A", ylab = "M", ...)
})

# codPlotDensity-method.
setGeneric("codPlotDensity", function(x, array, ...)
	standardGeneric("codPlotDensity"))
setMethod("codPlotDensity", "CodelinkSet",
function(x, array, lwd, ...)
{
	X <- getInt(x)
	islog <- getInfo(x, "log")

	if(missing(array))
		array <- 1:dim(X)[2]
	
	if(!islog) X <- log2(X)

	narray <- length(array)
	col <- rainbow(narray)

	d <- list()
	xlim <- c()
	ylim <- c()
	for(n in 1:narray) {
		k=array[n]
		d[[n]] <- density(X[, k], na.rm = TRUE)
		xlim <- c(xlim, d[[n]]$x)
		ylim <- c(ylim, d[[n]]$y)
	}

	xlim <- range(xlim, na.rm = TRUE)
	ylim <- range(ylim, na.rm = TRUE)

	plot(0, col = "white", xlim = xlim, ylim = ylim, xlab = "Intensity",
		ylab = "Density")
	
	if(missing(lwd)) lwd <- 2

	for(n in 1:narray)
		lines(d[[n]], col = col[n], lwd = lwd, ...)
	
	legend("topright", legend = sampleNames(x[,array]), fill = col, bty = "n")
	
	invisible(d)
})

setGeneric("codPlotScatter", function(x, array = 1, ...)
			standardGeneric("codPlotScatter"))
setMethod("codPlotScatter", "CodelinkSet",
function(x, array = 1, array2, label = "type", cutoff = c(-1, 1),
		snr.cutoff = 1,	legend.x, pch = ".", ...)
{
	#object, x=1, y=2, cutoff=FALSE, label="type", title=NULL, xlim=NULL, ylim=NULL
	label <- match.arg(label, c("type", "snr", "none"))
	type <- probeTypes(x)
	snr <- meanSNR(x)
	samples <- sampleNames(x)

	X <- getInt(x)
	islog <- getInfo(x, "log")
	
	X1 <- X[, array]
	xlab = samples[array]
	ylab = "Median"
	if(missing(array2)) {
		# compute mean in non-log scale and restore it if needed.
		X2 <- rowMeans(if(!islog) X else 2^X, na.rm = TRUE)
		if(islog) X2 <- log2(X2)
	} else {
		X2 <- X[, array2]
		ylab = samples[array2] 
	}
	
	if(!islog) {
		X1 <- log2(X1)
		X2 <- log2(X2)
	}
	
	if(missing(array2)) {
		title <- paste("Median vs", samples[array])
	} else {
		title <- paste(samples[array2], "vs", samples[array])
	}
	
	plotxy(X1, X2, label = label,	cutoff = cutoff, snr.cutoff = snr.cutoff,
			legend.x = legend.x, pch = pch, type = type, snr = snr, title = title, loess = FALSE, xyline = TRUE, 
			xlab = xlab, ylab = ylab, ...)
})

# codPlotImage-method.
setGeneric("codPlotImage", function(x, array = 1, ...)
	standardGeneric("codPlotImage"))
setMethod("codPlotImage", "CodelinkSet",
function(x, array = 1, signal = "bg", low = "white", high = "blue",
	log.it = FALSE, ...)
{
	signal <- match.arg(signal, c("bg", "intensity", "snr"))
	switch(signal,
		bg = X <- getBkg(x)[, array],
		intensity = X <- getInt(x)[, array],
		snr = X <- getSNR(x)[, array]
	)

	if(log.it) X <- log2(X)
	
	d <- getChipDimensions(x)
	if(is.null(d)) {
		gc <- 1
		gr <- 1
	} else {
		gc <- d["gc"]
		gr <- d["gr"]
	}

	rows <- logicalRow(x)
	cols <- logicalCol(x)

	or <- max(rows)
	oc <- max(cols)
	o <- matrix(NA, nrow = or, ncol = oc)
	for(n in 1:length(rows)) {
		o[rows[n], cols[n]] <- X[n]
	}

	col <- colorRampPalette(c(low, high))(123)
	op <- par(mar = c(1, 1, 1, 1))
	on.exit(par(op))

	if(is.null(d)) {
		sc <- oc/gc
		sr <- or/gr
	} else {
		sc <- d["sc"]
		sr <- d["sr"]
	}

	image(0:(gr * sr), 0:(gc * sc), o, col = col, xaxt = "n", yaxt = "n", ...)

	for (igrid in 0:gc) lines(c(0, gr * sr), rep(igrid * sc, 2), lwd = 2)
	for (igrid in 0:gr) lines(rep(igrid * sr, 2), c(0, gc * sc), lwd = 2)
	box(lwd = 2)

	mtext(paste("Array:", array, "Signal: ", signal, " Sample:",
		sampleNames(x)[array]), side=1, cex=0.8)
})
#
setGeneric("getChipDimensions", function(x)
	standardGeneric("getChipDimensions"))
setMethod("getChipDimensions", "character",
function(x)
{
	#if(missing(x)) return(NULL)
	switch(x,
		hwgcod = {
			gc <- 1
			gr <- 12
			sc <- 112
			sr <- 42
		},
		mwgcod = {
			gc <- 1
			gr <- 10
			sc <- 112
			sr <- 41
		},
		rwgcod = {
			gc <- 1
			gr <- 8
			sc <- 112
			sr <- 41
		},
		h20kcod = {
			gc <- 1
			gr <- 1
			sc <- 71
			sr <- 332
		},
		return(NULL)
	)
	return(c(gc = gc, gr = gr, sc = sc, sr = sr))
})
#
setMethod("getChipDimensions", "CodelinkSet",
function(x)
{
	getChipDimensions(x@annotation)
})
#
setGeneric("chipDevice", function(x, f = 3) standardGeneric("chipDevice"))
setMethod("chipDevice", "CodelinkSet",
function(x, f = 3)
{
	arrayNew(f = f, chip = x@annotation)
})

setGeneric("writeCodelink", function(object, file, dec = ".", sep = "\t", flag = FALSE, chip) standardGeneric("writeCodelink"))
setMethod("writeCodelink", "CodelinkSet",
function(object, file, dec = ".", sep = "\t", flag = FALSE, chip)
{
    if(missing(chip)) chip <- annotation(object)
    if(chip == "") stop("chip name is needed.")
    probes <- probeNames(object)
    probes.acc <- lookUp(probes, chip, "ACCNUM", load = TRUE)
    probes.eg <- lookUp(probes, chip, "ENTREZID")

	tmp <- cbind(probeNames(object), unlist(probes.acc), unlist(probes.eg), exprs(object), getSNR(object))	
    head <- c("PROBE_NAME", "ACCESSION", "ENTREZID", paste("INTENSITY_", sampleNames(object), sep = ""),
            paste("SNR_", sampleNames(object), sep = ""))
    
	if(flag) {
        tmp <- cbind(tmp, getFlag(object))
        head <- c(head, paste("FLAG_", sampleNames(object), sep = ""))
    }
    
	rownames(tmp) <- featureNames(object)
    tmp <- rbind(Index=head, tmp)
    write.table(tmp, file = file, quote = FALSE, sep = sep, dec = dec, col.names = FALSE)
})

setGeneric("averageProbes", function(object, parallel = FALSE)
			standardGeneric("averageProbes"))
setMethod("averageProbes", "CodelinkSet",
function(object, parallel = FALSE)
{
	uprobes <- unique(probeNames(object))
	mat <- exprs(object)
	samples <- sampleNames(object)
	
	mylapply <- lapply
	if (parallel) {
		if (is.loaded("mclapply", PACKAGE="parallel")) {
			mylapply <- get('mclapply', envir=getNamespace('multicore'))
		} else {
			message("Parallel functionality is provided by package 'parallel', please load the package to use it")
			message("Package 'parallel' is available by default in R>= 2.14.0")
		}
	}
	
	tmp <- mylapply(uprobes, function(x) {
		idx <- probeNames(object) == x
		int.m <- colMeans(mat[idx,, drop = FALSE], na.rm = TRUE)
		int.sd <- apply(mat[idx,, drop = FALSE], 2, function(x) sd(x, na.rm = TRUE))
		c(which(idx)[1], int.m, int.sd)
	})
	
	int.m <- matrix(unlist(tmp), ncol = ncol(object)*2 + 1, nrow = length(uprobes), byrow = TRUE)
	uidx <- int.m[, 1]
	int.m <- int.m[, -1]
	sel <- c(rep(TRUE, ncol(object)), rep(FALSE, ncol(object)))
	int.sd <- int.m[, !sel]
	int.m <- int.m[, sel]
	
	fdata <- new("AnnotatedDataFrame",
			data = pData(featureData(object))[uidx, c("probeName", "probeType")],
			varMetadata = varMetadata(featureData(object))[c("probeName", "probeType"), , drop = FALSE])
	featureNames(fdata) <- uprobes
	new("CodelinkSetUnique", exprs = int.m, sd = int.sd, phenoData = phenoData(object), featureData = fdata, annotation = annotation(object))
})

%\VignetteEngine{knitr}
%\VignetteIndexEntry{Codelink Intruction}
%\VignetteKeywords{Preprocessing, Codelink}
%\VignetteDepends{codelink}
%\VignetteDepends{knitr}
%\VignettePackage{codelink}

\documentclass{article}
\usepackage[utf8]{inputenc}
\usepackage[english]{babel}
\usepackage{geometry}
\geometry{verbose,tmargin=3cm,bmargin=3cm,lmargin=3cm,rmargin=3cm}

\newcommand{\Robject}[1]{{\texttt{#1}}}
\newcommand{\Rfunction}[1]{{\texttt{#1}}}
\newcommand{\Rpackage}[1]{{\texttt{#1}}}

\begin{document}

<<include=FALSE,cache=FALSE>>=
library(codelink)
library(knitr)
opts_chunk$set(fig.align = 'center', concordance=TRUE,width=50,fig.show="hold", tidy = TRUE, comment = "", highlight = FALSE, prompt = TRUE)
knit_hooks$set(no.mar = function(before, options, envir) {
    if (before) par(mar = rep(0,4))  # no margins.
})
@


\title{Introduction to the Codelink package}
\author{Diego Diez}
\maketitle

\section{Introduction}

This package implements methods to facilitate the preprocessing and analysis of Codelink microarrays. Codelink is a microarray platform for the analysis of gene expression that uses 30 base long oligonucleotides. Codelink is currently owned by Applied Microarrays, Inc. (previously was GE Healthcare and before that Amersham). There is a proprietary software for reading scanned images, perform spot intensity quantification and diagnostics. A Codelink microarray consists of a number of species-specific probes to measure gene expression, as well as some other control probes (see Table~\ref{tab:Type}). The Codelink software assigns quality flags to each spot (see Table~\ref{tab:Flag}) on the basis of a signal to noise ratio (SNR) computation (Eq: \ref{eq:SNR}) and other morphological characteristics as irregular shape of the spots, saturation of the signal or manufacturer spots removed. By default, the software performs background correction (subtract) followed by median normalization. The results can be exported in several formats as XML, Excel, plain text, etc.

The codelink package enables loading \emph{Codelink} data into R, and stores it as a CodelinkSet object. The \texttt{CodelinkSet-class} inherits from \texttt{ExpressionSet} all methods, and enables straightforward interfacing with other Bioconductor structures, and packages.

NOTE: the old \Robject{Codelink-class} infrastructure is maintained for backbard compatibility, and information about its use can be found in the vignette Codelink\_Legacy.pdf.


\begin{table}[ht]
\begin{center}
\begin{tabular}{r|l|c}
\hline \em Probe type & \em Description & \em Default weight\\
\hline
DISCOVERY & Measure gene expression & 1\\
POSITIVE & Positive control & 0\\
NEGATIVE & Negative control & 0\\
FIDUCIAL & Grid alignment & 0\\
OTHER & Other controls and housekeeping genes & 0\\
\hline
\end{tabular}
\caption{Probe types for Codelink arrays.}
\label{tab:Type}
\end{center}
\end{table}

\begin{table}[ht]
\begin{center}
\begin{tabular}{c|l|c|c}
\hline \em Flag & \em Description & \em Default value set & \em Default weight\\
\hline
G & Good signal (SNR $\geq$ 1) & & 1\\
L & Limit signal (SNR $<$ 1) & & 1\\
S & Saturated signal & & 1\\
I & Irregular shape & & 0\\
M & MSR spot (-9999) & NA & 0\\
C & Background contaminated & & 0\\
X & User excluded spots & & 0\\
\hline
\end{tabular}
\caption{Quality Flag description. SNR: Signal to Noise Ratio.}
\label{tab:Flag}
\end{center}
\end{table}

\begin{equation}
SNR=\frac{Smean}{(Bmedian + 1.5 * Bstdev)}
\label{eq:SNR}
\end{equation}



\section{Reading data}

Only Codelink data exported as plain text from the Codelink software is supported. Unfortunately the Codelink exported text format can have arbitrary columns and  header fields so depending of what has been exported, reading it into a \Robject{CodelinkSet} object may be more or less complicated. As a rule of thumb it is recommended to include in the exported files at least Spot\_mean and Bkgd\_median values so that background correction and normalization can be performed within R. In addition, Bkgd\_stdev will be needed to compute the SNR. If Raw\_intensity or Normalized\_intensity columns are present then it is possible to avoid background correction and/or normalization, and use the ones performed by the Codelink software. The Feature\_id column will be use to assign unique identifiers to each spot, so that CodelinkSet object can be read appropriately (or else will try to guess those). To read codelink data:

<<eval=FALSE>>=
# NOT RUN #
library(codelink)
# to read data as CodelinkSet object:
f = list.files(pattern="TXT")
codset = readCodelinkSet(filename=f)
# NOT RUN #
@

This assumes that the files have the extension "TXT" (uppercase) and are in the working directory. You can prepare a targets file with each file's name and additional phenotypic information, then pass this information to readCodelinkSet() so that it is stored in the CodelinkSet object.

<<eval=FALSE>>=
# NOT RUN #
pdata=read.AnnotatedDataFrame("targets.txt")
codset=readCodelinkSet(filename=pdata$FileName, phenoData=pdata)
# NOT RUN #
@

<<>>=
# sample dataset.
data(codset)
codset
@

To convert old Codelink objects into the new CodelinkSet the handy function Codelink2CodelinkSet() can be used:

<<>>=
data(codelink.example)
print(is(codelink.example))
tmp=Codelink2CodelinkSet(codelink.example)
tmp
@

\subsection{Flags and weights}
Traditionally the codelink package has used flag information to assign NAs to values. This behavior has been changed since the version released with Bioconductor 2.13 (October, 2013). To reproduce the old behavior call \texttt{readCodelinkSet()} with argument \texttt{old=TRUE}.

In the current implementation, only probes flagged as MSR spots (flag 'M'- which have an intensity value assigned of -9999), will be automatically converted to NA. This value cannot be adjusted since the value of the probes itself does not represent any measure of signal.

In addition to this, probe weights will be computed by default, based on the conversion table shown in tables \ref{tab:Type} and \ref{tab:Flag}. The weight computation follows this process. First, weights are assigned based on type, with DISCOVERY probes being assigned \texttt{weight=1} and other probes \texttt{weight=0}. Then, weights are adjusted based on flags. The worst weight (type or flag weights when multiple) is assigned to each probe. The weights assigned can be controled by the \texttt{type.weights} and \texttt{flag.weights} argument to \texttt{readCodelinkSet()}. It is possible also to reassign weights after reading with the function \texttt{createWeights()}. Weights can be used during preprocessing (background correction and normalization) and linear modeling.

<<>>=
w = createWeights(codset)
## NOTE: a proper replacement function will be provided later:
assayDataElement(codset,"weight")=w
@

\subsection{Accessing data}

Data stored in a CodelinkSet object can be accessed using several accessor functions:
<<>>=
# get signal intensities. alias: getInt()
head(exprs(codset))

# get background intensities.
head(getBkg(codset))

# get SNR values.
head(getSNR(codset))

# get flags.
head(getFlag(codset))

# get weights.
head(getWeight(codset))

# get phenoData:
head(pData(codset))

@

\section{Background correction}

If Spot\_mean and Bkgd\_median values are available then background correction can be performed with \texttt{codCorrect()}. Background correction methods are borrowed from the limma package, including methods \emph{none}, \emph{subtract}, \emph{half} and \emph{normexp}. The default is set to \emph{half}, because it is very fast. However, more sensitive (although slower) methods like \emph{normexp} are recommended. It is possible to assign an offset to avoid low intensity probes to have high M variances.

<<>>=
codset = codCorrect(codset, method = "half", offset = 0)
@

\section{Normalization}

Normalization of the background corrected intensities is done by the wrapper 
function \Rfunction{normalize} (or the alias \texttt{codNormalize()}). Here again, normalization is borrowed from the limma package. Methods \emph{median}, \emph{quantile} (the default) and \emph{loess} are available. 

<<>>=
codset = codNormalize(codset, method = "quantile")
@

Method \emph{loess} performs CyclicLoess normalzation and accepts weights. Weights are used in a per-probe fashion (that is, one weight for one probe, not different weights for each sample). When weights are used for normalzation the minimum of all the weights for each probe along all the samples will be used. This is to ensure that for each array there is an equal contribution of each probe in the normalization process.

<<eval=FALSE>>=
# NOT RUN
codset = codNormalize(codset, method = "loess", weights=getWeight(codset), loess.method="fast")
# NOT RUN
@

\section{Diagnostic plots}

There are some plot facilities to help diagnose the effect of background correction and normalization, as well as identify putative faulty arrays. The most commonly used functions are MA plots, density plots and array images. All these functions can be accessed through the function codPlot(). The parameter \texttt{what} specifies the type of plot: \emph{ma} (default), \emph{density}, \emph{scatter} and \emph{image} are valid choices. Figures 1 and 2 below show examples of these plotting functions.


<<fig.cap="MA plot (left) and density plot (right).",fig.height=4,fig.width=4,out.width='.49\\linewidth'>>=
codPlot(codset) # by default MA plot.
codPlot(codset, what="density")
@


When the columns \emph{Logical\_row} and \emph{Logical\_col} are present in the original data files, this information is used to assign the physical location of each probe in the array to plot a pseudo image. It is possible to plot the background intensities (default), the spot mean, raw and normalized intensities and the SNR values. This images are useful to identify spatial artifact that may be affecting the analysis.

<<fig.cap="Pseudo image plot of an array",fig.height=1,fig.width=3>>=
codPlot(codset, what="image")
@


\section{Fitting linear models}

A typical analysis include the testing for differentially expressed probes between two populations. This can be performed using a variety of different R/Bioconductor packages, but the \texttt{limma} package is one of the most popular options. Limma can readily use \texttt{CodelinkSet} objects, and can take advantage of weights generated during data reading. In this case, weights will be use probe-wise (i.e. different weights for the same probe in different samples will be considered).

<<>>=
fit = lmFit(codset, design=c(1,1,2,2), weights=getWeight(codset))
fit2 = eBayes(fit)
topTable(fit2)
@

\section{Citation}

<<>>=
citation(package="codelink")
@

\section{Session info}

<<>>=
sessionInfo()
@

\end{document}

%\VignetteEngine{knitr}
%\VignetteIndexEntry{Codelink Legacy}
%\VignetteKeywords{Preprocessing, Codelink}
%\VignetteDepends{codelink}
%\VignetteDepends{knitr}
%\VignettePackage{codelink}

\documentclass{article}
\usepackage[utf8]{inputenc}
\usepackage[english]{babel}
\usepackage{geometry}
\geometry{verbose,tmargin=3cm,bmargin=3cm,lmargin=3cm,rmargin=3cm}

\newcommand{\Robject}[1]{{\texttt{#1}}}
\newcommand{\Rfunction}[1]{{\texttt{#1}}}
\newcommand{\Rpackage}[1]{{\texttt{#1}}}

\begin{document}


<<include=FALSE,cache=FALSE>>=
library(codelink)
library(knitr)
opts_chunk$set(fig.align = 'center', concordance=TRUE)
@

\title{Codelink Legacy: the old Codelink class}
\author{Diego Diez}
\maketitle

\section{Introduction}

Codelink is a platform  for the analysis of gene expression on biological 
samples property of Applied Microarrays, Inc. (previously was GE Healthcare and 
Amersham). The hybridization reagents are still supplied by GE Healthcare.

The system uses 30 base long oligonucleotide probes for expression testing.  
There is a proprietary software for reading scanned images, doing spot 
intensity quantization and some diagnostics. The software assigns quality flags 
(see Table~\ref{tab:Flag}) to each spot on the basis of a signal to noise ratio 
(SNR) computation (Eq: \ref{eq:SNR}) and other morphological characteristics as 
irregular shape of the spots, saturation of the signal or manufacturer spots 
removed. By default, the software performs background correction (subtract) 
followed by median normalization. The results can be exported in several 
formats as XML, Excel, plain text, etc. This library allows to read Codelink 
plain text exported data into R \cite{r} for the analysis of gene expression 
with any of the available tools in R+Bioconductor\cite{BIOC}.

For storing the available data, a new class \Robject{Codelink} was designed.
The now defunt \Robject{exprsSet} in the \Rpackage{Biobase} package was not a 
convenient store class for Codelink data, because it didn't allow probes with 
duplicated names, as those names are stored as rownames in the expression 
matrix.

Currently there is experimental support for a \Robject{ExpressionSet} derived 
class that can accomodate duplicated probe names in an 
\Robject{AnnotatedDataFrame} whereas the feature ids would be used for row 
names when available.

\begin{table}[ht]
\begin{center}
\begin{tabular}{|c|l|}
\hline \em Flag & \em Description \\
\hline
G & Good signal (SNR >= 1)\\
L & Limit signal (SNR < 1)\\
I & Irregular shape \\
S & Saturated signal \\
M & MSR spot \\
C & Background contaminated\\
X & User excluded spots\\
\hline
\end{tabular}
\caption{Quality Flag description. SNR: Signal to Noise Ratio.}
\label{tab:Flag}
\end{center}
\end{table}

\begin{equation}
SNR=\frac{Smean}{(Bmedian + 1.5 * Bstdev)}
\label{eq:SNR}
\end{equation}

\begin{table}[ht]
\begin{center}
\begin{tabular}{|r|l|}
\hline \em Probe type & \em Description \\
\hline
DISCOVERY & Gene expression testing probes \\
POSITIVE & Positive control probes \\
NEGATIVE & Negative control probes \\
FIDUCIAL & Grid alignment probes \\
OTHER & Other controls and housekeeping gene probes \\
\hline
\end{tabular}
\caption{Probe types for Codelink arrays.}
\label{tab:Type}
\end{center}
\end{table}

\section{Reading data}

Currently only data exported as plain text from Codelink software is supported.  
Unfortunately the Codelink exported text format can have arbitrary columns and 
header fields so depending of what you have exported you can read it or not.  
The suggestion is that you put on the files everything you can, including  
Spot\_mean and Bkgd\_median values so you can do background correction and 
normalization in R. In addition, Bkgd\_stdev is needed to compute the SNR. If 
you put Raw\_intensity or Normalized\_intensity columns then you can also read 
it directly and avoid background correction and/or normalization but this is 
not recommended. To read some Codelink files you do:

<<input, eval=FALSE>>=
# NOT RUN #
library(codelink)
foo <- readCodelink()
summaryFlag(foo) # will show a summary of flag values.
# NOT RUN #
@

This suppose that your files have the extension ``TXT'' (uppercase)
and that they are in your working directory. If this is not the case you can 
specify the files to be read with the 'file' argument. The function 
\Rfunction{readCodelink} returns and object of \Robject{Codelink} similar to 
that:

<<data, comment=NA>>=
data(codelink.example)

codelink.example
@

\begin{table}[ht]
\begin{center}
\begin{tabular}{|r|l|}
\hline \em Slot & \em Description \\
\hline
product & Chip name description \\
sample & Sample names vector \\
file & File names vector \\
name & Probe names vector \\
type & Probe types vector \\
method & Methods applied to data \\
method\$background & Background correction method used \\
method\$normalization & Normalization method used \\
method\$merge & Merge method used \\
method\$log & Logical: If data is in log scale \\
flag & Quality flag matrix \\
Smean & Mean signal intensity matrix \\
Bmedian & Median background intensity matrix \\
Ri & Raw intensity matrix \\
Ni & Normalized intensity matrix \\
snr & Signal to Noise Ratio matrix \\
cv & Coefficient of Variation matrix \\
\hline
\end{tabular}
\caption{Description of Codelink object slots.}
\label{tab:Slots}
\end{center}
\end{table}

The \Robject{Codelink-class} is basically a list that stores information in 
several slots (see Table~\ref{tab:Slots}). The chip type (product slot) is read 
from the PRODUCT field if found in the header of Codelink files. If it is not 
found then a warning message is shown and product slot is set to "Unknown". If 
the product is not the same in all the files the reading is canceled with an 
error message.

By default, all spots flagged with M, I, and S flags are set to NA. This can be
controlled with the flag argument of \Rfunction{readCodelink}. The flag 
argument is a list that can contain a valid flag identifier and a value for 
that flag.  For example, if you want to set all M flagged spots to 0.01 and let 
other spot untouched you do:


<<input-flag, eval=FALSE>>=
foo <- readCodelink(flag = list(M = 0.01) )
@

It is possible to find probes wit more that one flag assigned, i.e. CL for a 
probe labeled as C and L, CLI for a probe labeled as C, L and I, and so on. A a 
regular expression is used to find flag types in an attempt to to manage all 
the possible situations. When two user modified flags fall in the same probe 
the more restricting (NA being the most) is assigned.

\section{Background correction}

If you have Spot\_mean values Bkgd\_median values the you can apply one of the 
several background correction methos interfaced. This is done by the function 
\Rfunction{bkgdCorrect}. To see the different options look at ?bkgdCorrect. For 
instance, if you want to apply \emph{half} method you do:


<<bkgcorrect, eval=FALSE>>=
foo <- bkgdCorrect(foo, method = "half")
@

The default method used is \emph{half} and is based in the same method applied 
in the \Rpackage{limma} \cite{limma} package to two channel microarrays. In 
this method, the median background intensity (Bmedian) is subtracted from mean 
spot intensity (Smean) and any value smaller that 0.5 is shift to 0.5 to ensure 
no negative numbers are obtained that would prevent to transform the data into 
log scale. Other available methods are \emph{none} that let the spot 
intensities untouched, \emph{subtract} that is analog to the default method 
used in the Codelink software and \emph{normexp} and interface to the method 
available in the \Rpackage{limma} package.

\section{Normalization}

Normalization of the background corrected intensities is done by the wrapper 
function \Rfunction{normalize}. The default method is \emph{quantile} 
normalization that in fact call \Rfunction{normalizeQuantiles()} from 
\Rpackage{limma} package (allowing for NAs). There is also the possibility to 
use a modified version of CyclicLoess from \Rpackage{affy} \cite{affy} package 
that allow using weights and missing values. Finally, the \emph{median} 
normalization allows to normalize using a method analog to the default method 
in the Codelink software. To normalize you usually do:

<<normalize, eval=FALSE>>=
foo <- normalize(foo, method = "quantiles")
@

By default, \Rfunction{normalize} return log2 intensity values.  This could be
controlled setting the parameter log.it to FALSE.

\section{Plotting}

There are some diagnostic plots available for the \Robject{Codelink} object.  
These are functions for producing MA plots (\Rfunction{plotMA}), scatterplots 
(\Rfunction{plotCorrelation}) and density plots (\Rfunction{plotDensities}).  
All functions use the available intensity value (i.e.  Smean, Ri or Ni) to make 
the plot.

The functions \Rfunction{plotMA} and \Rfunction{plotCorrelation} can highlight 
points based on the Spot Type, which is the default behavior or using the SNR 
values.  The mode is controlled with argument \emph{label}.  
\Rfunction{plotCorrelation} requires arguments \emph{array1} and \emph{array2} 
to be set in order to select which arrays are going to be plotted. For 
\Rfunction{plotMA} if only \emph{array1} is specified, the values are plotted 
againts a pseudoarray constructed with the mean of the probe intensities along 
all available arrays. M and A values are computed following equations 
\ref{eq:M} and \ref{eq:A}.


\begin{equation}
M=Array2-Array1
\label{eq:M}
\end{equation}

\begin{equation}
A=\frac{Array2+Array1}{2}
\label{eq:A}
\end{equation}

<<plotMA>>=
plotMA(codelink.example)
@

The function \Rfunction{plotDensities} plot the density of intensity values of
all arrays. It can plot only a subset of arrays if the \emph{subset} argument 
is supplied.

<<plotDensity>>=
plotDensities(codelink.example)
@

When Logical\_row and Logical\_col columns are exported into they are stored 
into the \emph{logical} slot. This information stores the physical location of 
each probe in the array, and can be used to plot pseudo images of the array 
intensities. To plot a pseudo image you should use:

<<image, eval=FALSE>>=
imageCodelink(foo)
imageCodelink(foo, what = "snr")
@

It is possible to plot the background intensities (default), the spot mean, 
raw and normalized intensities and the SNR values. This images are useful to 
identify spatial artifact that may be affecting the analysis.

\section{Miscellaneous}

There are also some miscellaneous functions used in some analysis that could be
useful for someone.

\subsection{Export to file}

The function \Rfunction{writeCodelink} exports a \Robject{Codelink} object to a file. The file will contain probe intensities and, if specified flag = TRUE, probe quality flags.


\subsection{Merging arrays}

In case you want to merge array intensities the \Rfunction{mergeArray} function 
help on this task. It computes the mean of Ni values on arrays of the same 
class.  The grouping is done by means of the \emph{class} argument (numerical 
vector of classes). New sample names should be assigned to the sample slot 
using the \emph{names} argument. The function also returns the coefficient of 
variation in the \emph{cv} slot. The distribution of coefficients of variations 
can be checked with the function \Rfunction{plotCV}.

<<merge, eval=FALSE>>=
foo <- mergeArray(foo, class = c(1, 1, 2, 2), names = c("A", "B"))
plotCV(foo)
@

\section{Problems reading data}
Some updated version of the Codelink software changed the order in which probes 
are printed in the exported text files. That makes files from these different 
software version impossible to be analyzed together. The function 
\Rfunction{readCodelink} have a new argument \emph{check} TRUE by default, that 
check the Probe\_name columns to see if they have the same order in all the 
arrays at the cost is a little extra loading time. This behaviour can be turn 
off by setting \emph{check} to FALSE. If an ordering problem is found, a 
warning message is print but the reading of data is not stopped, allowing the 
visual examination of the data. In this case, the best option is to export the 
old files again using the updated version of the software. If this is 
impossible for whatever reason, the codelink package can try to fix the order 
of the files.

To the aim, a new argument \emph{fix = TRUE} can be passed to 
\Rfunction{readCodelink}. The method will try to order the probes using the 
Feature\_id column, which is a combination of Logical\_row and Logical\_col and 
for this, a unique identifier of each probe. This is the optimal fix, and the 
new data should be perfectly ordered. If that column is missing, the it will 
try to order the data using the Probe\_name column. This is a sub-optimal 
solution because as the fiducial, control and some discovery probes have 
duplicated Probe\_name, they may be end messed up. In this case, the best 
solution again, is to try to export again the data with the updated version of 
the software.


<<read-fix, eval=FALSE>>=
foo <- readCodelink(fix = TRUE)
@

\section{Future improvements}
As the new classes in the Biobase package (\Robject{eSet} and 
\Robject{ExpressionSet}) allow more flexible data structures, a 
reimplementation of the \Robject{Codelink-class} based on 
\Robject{ExpressionSet-class} will be the next major feature added to this 
package. This will allow a better integration with other tools available 
through the Bioconductor project.

Right now, there is some experimental implementation of the new codebase. You 
need to use the wrapper function \Rfunction{readCodelinkSet}.

<<read-codelinkSet, eval=FALSE>>=
foo <- readCodelinkSet()
@


\bibliographystyle{plain}
\bibliography{codelink}


\end{document}

\name{plotMA}
\alias{plotMA}
\title{MA plot}
\description{
	Takes a Codelink object and plot M vs A.
}
\usage{
  plotMA(object, array1 = 1, array2 = NULL, cutoff = c(-1, 1), label = NULL,
    type = NULL, high.list = NULL, high.col = "blue", high.pch = 21,
	high.bg = "cyan",	snr = NULL, snr.cutoff = 1, legend.x = NULL, pch = ".", 
	subset = NULL, title = NULL, xlim = NULL, ylim = NULL)
}
\arguments{
  \item{object}{an object of class "Codelink" or "MAarrayLM".}
  \item{array1}{first array to be used.}
  \item{array2}{second array to be used.}
  \item{cutoff}{cutoff to be used as fold change markeer.}
  \item{label}{type of labeling used in legend.}
  \item{type}{spot type information.}
  \item{high.list}{list of genes highlighted.}
  \item{high.col}{color used for high genes.}
  \item{high.pch}{pch used for high genes.}
  \item{high.bg}{background color used for high genes.}
  \item{snr}{vector with SNR values, usually, taking rowMeans() from a SNR matrix.}
  \item{snr.cutoff}{SNR cutoff used for label spots.}
  \item{legend.x}{relative position of the legend.}
  \item{pch}{pch style used to main spots.}
  \item{subset}{subset of spots used to plot based on 'type' slot.}
  \item{title}{title of the plot.}
  \item{xlim}{range for the X axis.}
  \item{ylim}{range for the Y axis.}
}
\details{

  This function has suffered recent re-working, to increase the usability and 
  to clean a little bit the code.

  If array2 is NULL a median array is computed using all available arrays.  
  Then the values of M and A are computed using the following formula:

                             M = array2 - array1

                          A = (array2 + array1) / 2

  If type information is available in the Codelink object, or provided throught
  the 'type' argument, spots are colored based on that. DISCOVERY spots are 
  plotted black with pch = "." whereas the other classes are plotted with 
  different background colors, using gray as border to increase contrasts. For 
  that pch = 21 is used. If snr is specified as label option, the SNR is used 
  to label spots, if available in the Codelink object. In this case, the mean 
  SNR across all arrays is used when array2 = NULL.

  Some parameters may not be working right now, as the new function is using a 
  different method to labels spots.

  The legend is 'automagically' located, but this can be overrided with the 
  legend.x argument.

  In addition, a subset of the spots can be plotted based on type information 
  when available. This allows, for example, to plot only DISCOVERY spots.
}
\author{Diego Diez}
\examples{
\dontrun{
    data(codelink.example)
    plotMA(codelink.example)
}
}
\keyword{documentation}
\keyword{utilities}

\name{decDetect}
\alias{decDetect}
\title{Determine decimal type of Codelink files}
\description{
	Determine decimal type of Codelink files.
}
\usage{
    decDetect(file, nlines)
}
\arguments{
    \item{file}{the file to be read.}
    \item{nlines}{number of lines to skip.}
}
\value{Decimal type.}
\author{Diego Diez}
\keyword{documentation}
\keyword{utilities}

\name{reportCodelink}
\alias{reportCodelink}
\title{Write a report of genes selected in HTML}
\description{
  Takes a list of genes as argument and writes an HTML page containing informa-
tion about these genes: Unigene, Genbank, Entrez Gene, etc.
}
\usage{
	reportCodelink(object, chip, filename = NULL, title = "Main title",
                 probe.type = FALSE, other = NULL, other.ord = NULL)
}
\arguments{
  \item{object}{an object of class "Codelink" or a list of genes.}
  \item{chip}{the chip description package.}
  \item{filename}{file name used in the report.}
  \item{title}{title used in the report.}
  \item{probe.type}{logical; if TRUE Probe type information is  writen.}
  \item{other}{list of vectors containing aditional values to add to the report.}
  \item{other.ord}{slot name in other to order genes by.}
}
\value{Nothing, only the HTML file generated.}
\author{Diego Diez}
\seealso{\code{\link[annotate]{htmlpage}}}
\keyword{documentation}
\keyword{utilities}

\name{plotCorrelation}
\alias{plotCorrelation}
\title{Plot correlation scatterplot between two arrays}
\description{
	Takes a Codelink object as argument and plot Correlation scatterplot
	of two arrays.
}
\usage{
	plotCorrelation(object, x=1, y=2, cutoff=FALSE, label="type", title=NULL, xlim=NULL, ylim=NULL)
}
\arguments{
	\item{object}{an object of class "Codelink".}
	\item{x}{array to be used in x axis.}
	\item{y}{array to be used in y axis.}
	\item{cutoff}{cutoff used to show fold change.}
	\item{label}{labels to shown.}
	\item{title}{The title of the plot.}
        \item{xlim}{range for the X axis.}
        \item{ylim}{range for the Y axis.}
}
\author{Diego Diez}
\seealso{\code{\link{plot}}}
\examples{
\dontrun{
    data(codelink.example)
    plotCorrelation(codelink.example)
}
}
\keyword{documentation}
\keyword{utilities}

\name{codset}
\docType{data}
\alias{codset}
\title{Dataset of class 'CodelinkSet'}
\description{
  Dataset from a h20kcod (Codelink Human UniSet I 20k) array containing 4
  samples and ~20000 probes.
}
\usage{data(codelink.example)}
\format{
	A \code{\link[codelink:CodelinkSet-class]{Codelink}} object containing 
20469 probes and 4 samples.
}
\author{Diego Diez}
\keyword{datasets}

\name{arraySize}
\alias{arraySize}
\title{Determine the size of the array}
\description{
  When loading Codelink arrays in text format (as exported from the Codelink 
  software) this function retrieves the correct size of the array. This is 
  useful beacuse those files contain and indetermined number of empty lines at 
  the end. Thus, reading the entire data matrix don't works.
}
\note{
  Not meant to be used directly.
}
\author{Diego Diez}
\keyword{documentation}
\keyword{utilities}

\name{logCodelink}
\alias{logCodelink}
\title{Application of logCodelink to Codelink object}
\description{
	Takes a Codelink object and apply logCodelink to intensity values.
}
\usage{
logCodelink(object)
}
\arguments{
  \item{object}{an object of class "Codelink" or a list of genes.}
}
\value{A Codelink object with logCodelink intensities.}
\author{Diego Diez}
\seealso{\code{\link{log2}}}
\examples{
\dontrun{
    data(codelink.example)
    codelink.example <- logCodelink(codelink.example)
}
}
\keyword{documentation}
\keyword{utilities}

\name{normalize}
\alias{normalize}
\alias{normalize,Codelink-method}
\alias{normalize,CodelinkSet-method}
\alias{codNormalize}
\alias{codNormalize,CodelinkSet-method}

\title{Normalization wrapper for Codelink/CodelinkSet objects.}
\description{
	Takes a CodelinkSet or Codelink object and applies normalization to intensity values.
}
\usage{

\S4method{normalize}{Codelink}(object, method="quantiles", log.it=TRUE, preserve=FALSE, weights=NULL, loess.method="fast")
	
\S4method{normalize}{CodelinkSet}(object, method="quantile", log.it=TRUE, weights=NULL, loess.method="fast")
    
\S4method{codNormalize}{CodelinkSet}(object, method="quantile", log.it=TRUE, weights=NULL, loess.method="fast")
}
\arguments{
  \item{object}{an object of class CodelinkSet or Codelink.}
  \item{method}{method to use in normalization.}
  \item{log.it}{logical; if data should be log2.}
  \item{preserve}{logical; if Ri slot should be preserved.}
  \item{weights}{weights for method CyclicLoess.}
  \item{loess.method}{loess method to be used, default to 'fast'.}
}
\value{A CodelinkSet or Codelink object with normalized intensity values.}
\details{
  Currently supported methods include "median", "quantile" and "loess". Normalization functions are borrowed from the limma package.
  
  If a matrix of weights is passed for CyclicLoess normalization, a vector is constructed taking the lowest weight for each probe.
}
\author{Diego Diez}
\examples{
	data(codset)
	# Background correction.
	codset <- codCorrect(codset, method = "half")
	# Normalization.
	codset <- normalize(codset, method = "quantile")
}
\keyword{documentation}
\keyword{utilities}

\name{codPlot}
\alias{codPlot}
\alias{codPlot,CodelinkSet-method}
\alias{codPlot,MArrayLM-method}

\title{
Diagnostic plots for CodelinkSet object.
}
\description{
This function performs several types of diagnostic plots using information from CodelinkSet objects.
}
\usage{
codPlot(object, array, what = "ma", ...)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{object}{a CodelinkSet object.}
  \item{array}{The array to use for plotting (by default array=1).}
  \item{what}{the type of plot (by default ma).}
  \item{\dots}{additional arguments passed to the appropriate functions.}
}
\details{

Plot types:

\describe{
    \item{\code{scatter}}{
    	Scatter plot of the selected array intensities vs. the median array, or a second array if specified.
    }
    \item{\code{ma}}{
    	MA plot of the selected array intensities vs. the median array, or a second array if specified. In the x-axis the value of A (mean abundance; A = Array1 + Array2) is plotted. In the y-axis the value of M (difference; M = Array1 - Array2) is plotted.
    }
    \item{\code{density}}{
    	The distribution of intensities for each array in the dataset is plotted as a kernel density.
    }
    \item{\code{image}}{
    	If the logical position of the probes in the array are available, this plot produces a pseudoimage of the array. The values of the "intensity", "bg" or "snr" can be selected with the 'signal' parameter (default signal="bg"). Color range can be controled with 'high' and 'low' parameters.
    }
	}
}
\author{
Diego Diez
}

\name{plotDensities}
\alias{plotDensities}
\alias{plotDensities,Codelink-method}
\title{Plot Densities}
\description{
	Takes a Codelink object and plot the distributions of intensities.
}
\usage{
plotDensities(object, what = NULL, title = NULL, col = NULL, 
	legend.title = NULL, legend.cex=1, ...)
}
\arguments{
  \item{object}{an object of class "Codelink"}
  \item{what}{what data to plot, one of "bg", "smean", "snr", "ri" or "ni"}
  \item{title}{title of the plot}
  \item{col}{vector of colors}
  \item{legend.title}{if provided, a title for the legend}
  \item{legend.cex}{font factor use in legend}
  \item{...}{further arguments to be passed to plot() and line() functions (e.g. lwd, etc.)}
}
\author{Diego Diez}
\examples{
\dontrun{
    data(codelink.example)
    plotDensities(codelink.example)
}
}
\keyword{documentation}
\keyword{utilities}

\name{writeCodelink}
\alias{writeCodelink}
\alias{writeCodelink,CodelinkSet-method}
\title{Write a Codelink object to file.}
\description{
 Export of the data from a codelink object to a text file.
}
\usage{
  writeCodelink(object, file, dec = ".", sep = "\t", flag = FALSE, chip)
}
\arguments{
  \item{object}{an object of class "Codelink".}
  \item{file}{filename to write object to.}
  \item{dec}{decimal character to use.}
  \item{sep}{delimiter character to use.}
  \item{flag}{should the Codelink flags be written.}
  \item{chip}{chip package to use, normally guessed.}
}
\details{
By default, intensities and SNR are wrote to the file. If set, the flag are 
also output. The header have  "INTENSITY\_", "SNR\_" and "FLAG\_" respectibely 
appendend to the sample name for those values. The default delimiter is the tab 
character, but that can be set with the sep argument. The default decimal 
character is the point.
}
\author{Diego Diez}
\examples{
\dontrun{
    data(codelink.example)
    writeCodelink(codelink.example, file = "foo.txt")
}
}
\keyword{documentation}
\keyword{utilities}

\name{CodelinkSet-class}
\alias{class::CodelinkSet}
\alias{CodelinkSet}
\alias{CodelinkSet-class}
\alias{codPreprocess}
\alias{codPreprocess,CodelinkSet-method}
\alias{getBkg}
\alias{getInt}
\alias{getInfo}
\alias{getSNR}
\alias{getFlag}
\alias{getWeight}
\alias{getBkg,CodelinkSet-method}
\alias{getInt,CodelinkSet-method}
\alias{getInfo,CodelinkSet-method}
\alias{getSNR,CodelinkSet-method}
\alias{getFlag,CodelinkSet-method}
\alias{getWeight,CodelinkSet-method}
\alias{exprs,CodelinkSet-method}
\alias{Codelink2CodelinkSet}


\alias{probeNames}
\alias{probeNames,CodelinkSet-method}
\alias{probeNames,CodelinkSetUnique-method}
\alias{probeTypes}
\alias{probeTypes,CodelinkSet-method}

\docType{class}

\title{Class CodelinkSet}
\description{This is the storage class for Codelink data}
\section{Objects from the Class}{
  Object are created after reading text codelink files with readCodelinkSet()}

\section{Description}{
	CodelinkSet objects are derived from ExpressionSet and therefore inherits all their methods. Additional methods are defined to extract information:

	Old Codelink-class objects can be converted into CodelinkSet instances using the function Codelink2CodelinkSet.
}

\section{Extends}{
   Directly extends class \code{\link{ExpressionSet}}.
}

\section{Methods}{
Class-specific methods:

	\describe{
    \item{\code{getWeight}}{Get the matrix of weights.}
	}
}

\author{Diego Diez}

\examples{
\dontrun{
    data(codset)
}
}
\keyword{classes}

\name{codelink.example}
\docType{data}
\alias{codelink.example}
\title{Dataset of class 'Codelink'}
\description{
  Dataset from a h20kcod (Codelink Human UniSet I 20k) array containing 2 
  samples and ~20000 probes.
}
\usage{data(codelink.example)}
\format{
	A \code{\link[codelink:Codelink-class]{Codelink}} object containing 20469 
	probes and 2 samples.
}
\author{Diego Diez}
\keyword{datasets}

\name{SNR}
\alias{SNR}
\title{Calculate SNR}
\description{
	Compute SNR inside read.Codelink.
}
\usage{
	SNR(Smean, Bmedian, Bstdev)
}
\arguments{
  \item{Smean}{matrix of Smean intensities.}
  \item{Bmedian}{matrix of background median intensities.}
  \item{Bstdev}{matrix of background standard deviation.}
}
\author{Diego Diez}
\keyword{documentation}
\keyword{utilities}

\name{plotCV}
\alias{plotCV}
\title{Plot of CV}
\description{
	Takes a Codelink object and plot de distribution of CV after applying 
	mergeCodelink.
}
\usage{
	plotCV(object, subset=c(1:dim(object)[2]), cutoff=NULL, title=NULL, legend.cex=1)
}
\arguments{
	\item{object}{an object of class "Codelink".}
	\item{subset}{subset of arrays to plot}
	\item{cutoff}{cutoff of CV to be shown.}
	\item{title}{title of the plot.}
	\item{legend.cex}{factor to apply to the fonts in the legend to fit.}
}
\author{Diego Diez}
\keyword{documentation}
\keyword{utilities}

\name{fc2Cond}
\alias{fc2Cond}
\title{Select probes based on fold change calculation}
\description{
  Takes a Codelink object and calculate fold changes (M) between two conditions 
(samples). Then select genes based on those who pass the passed cutoff.
}
\usage{
  fc2Cond(object, cond1=NULL, cond2=NULL, fc=1.0, verbose=FALSE)
}
\arguments{
  \item{object}{an object of class "Codelink".}
  \item{cond1}{numeric or character; First condition to compute M.}
  \item{cond2}{numeric or character; Second condition to compute M.}
  \item{fc}{value of the fold change cutoff}
  \item{verbose}{logical; if some information is printed on the console.}
}
\value{A logical vector indicating which genes pass the cutoff}
\details{
  Conditions can be passed as characters or as numeric index from the sample 
slot. The intensities are internally transformed to log2 if needed. The M value 
is computed as:

   M = cond1 - cond2
}
\author{Diego Diez}
\keyword{documentation}
\keyword{utilities}

\name{Nondocumented-objects}
\alias{Nondocumented-objects}

\alias{annotation,Codelink-method}

% generic.
\alias{meanSNR}
\alias{chipDevice}
\alias{filterSNR}
\alias{codUniverse}

\alias{writeCluster}
\alias{writeCluster,ExpressionSet-method}
\alias{writeCluster,MArrayLM-method}
\alias{writeCluster,Codelink-method}
\alias{writeCluster,CodelinkSet-method}

% matrix-methods.
\alias{filterSNR,matrix-method}

% Codelink-methods.
\alias{filterSNR,Codelink-method}
\alias{codUniverse,Codelink-method}

% CodelinkSet-methods.
\alias{meanSNR,CodelinkSet-method}
\alias{chipDevice,CodelinkSet-method}
\alias{filterSNR,CodelinkSet-method}
\alias{codUniverse,CodelinkSet-method}

% non-generic functions.
\alias{Codelink2eSet}
\alias{eSet2Codelink}
\alias{guessAnnotation}
\alias{summary,Codelink-method}
\alias{summaryFlag}
\alias{summaryFlag,Codelink-method}
\alias{summaryFlag,matrix-method}


\title{Here goes every undocumented function.}
\description{
Every function without help page is redirected here. Either because it is not 
intended to have a help page or because it is a bug. If you think it is the 
later please contact the package maintainer.

NOTE: Actually it is none of that reasons but the fact that the package isunder 
heavy development.
}

\author{Diego Diez}
\keyword{documentation}
\keyword{internal}

\name{bkgdCorrect}
\alias{bkgdCorrect}
\alias{codCorrect}
\alias{codCorrect,CodelinkSet-method}

\title{Background correction of intensity values.}
\description{
Takes a CodelinkSet or Codelink object with Spot mean and Bkgd median values and performes background correction using one of the methods available.
}
\usage{
bkgdCorrect(object, method = "half", preserve = FALSE, verbose = FALSE, offset = 0)

\S4method{codCorrect}{CodelinkSet}(object, method = "half", normexp.method="saddle", offset=0)
}
\arguments{
  \item{object}{an object of class CodelinkSet or Codelink.}
  \item{method}{the correction method to use, one of "none", "subtract", "half" and "normexp".}
  \item{preserve}{logical; if Smean and Bmedian slots should be preserved.}
  \item{verbose}{logical; if TRUE print some information with method normexp.}
  \item{normexp.method}{method used with normexp correction.}
  \item{offset}{numeric; value to add to intensities.}
}
\value{An object of class Codelink with corrected intensity values, that is
	Ri slot.}
\details{
  Available methods are:
  . none: left intensities untouched.
  . subtract: simple subtraction of Bkgd median from Spot mean.
  . half: the same as above but aboid negative values setting all intensity
          values below zero to 0.5.
  . normexp: apply normexp background adjustment from package limma.
}
\author{Diego Diez}
\examples{
    data(codset)
    codset <- codCorrect(codset, method = "half")
}
\keyword{documentation}
\keyword{utilities}

\name{as.matrix.Codelink}
\alias{as.matrix.Codelink}
\title{Return a matrix of intensity values}
\description{
	Takes a Codelink object and returns a matrix with
	the intensity values available.
}
\usage{
  \method{as.matrix}{Codelink}(x, ...)
}
\arguments{
  \item{x}{an object of class "Codelink".}
  \item{...}{additional arguments added to generic as.matrix since R-2.5.x}
}
\value{A matrix with the intensity values.}
\author{Diego Diez}
\seealso{\code{\link{as.matrix}}}
\examples{
\dontrun{
	data(codelink.example)
	mat <- as.matrix(codelink.example)
	is(mat)
}
}
\keyword{documentation}
\keyword{utilities}

\name{readCodelink}
\alias{readCodelink}
\alias{readCodelinkSet}
\title{Read Codelink Bioarrays Data}
\description{
  Read data exported as text by Codelink Software. It reads values (normalized
by Codelink Software or not) flags and information about probes.
}
\usage{
readCodelink(files = list.files(pattern = "TXT"), sample.name =
                 NULL, flag, flag.weights, type.weights, dec = NULL,
                 type = "Spot", preserve = FALSE, verbose = 2,
                 file.type = "Codelink", check = TRUE, fix = FALSE, old
                 = FALSE)
readCodelinkSet(filename, path, phenoData = NULL, \dots)
}
\arguments{
  \item{files}{list of files to read.}
  \item{sample.name}{vector of same length as files with sample names.}
  \item{flag}{list with values to assign based on Flag quality values.}
  \item{flag.weights}{weights assigned to each probe flag.}
  \item{type.weights}{weights assigned to each probe type.}
  \item{dec}{character indicating the decimal character used in the files.}
  \item{type}{character indicating which base value to read from files.}
  \item{preserve}{logical, if TRUE Bkgd\_stdev slot is not removed (if present).}
  \item{verbose}{numerical, set the level of information. Level 2 set as old behaviour. Level > 2 output some debug info.}
  \item{file.type}{exported file type, currently Codelink or XLS file formats supported.}
  \item{check}{logical, check for probe order consistency.}
  \item{fix}{logical, try to fix probe order consistency.}
  \item{old}{logical, whether NA is assigned to intensities based on flags.}

  \item{filename}{list of files.}
  \item{path}{path to the files.}
  \item{phenoData}{phenoData object with sample information.}
  \item{\dots}{further arguments to be passed.}
}
\value{readCodelink returns an object of class "Codelink", whereas readCodelinkSet() returns a CodelinkSet object.}
\author{Diego Diez}
\examples{
\dontrun{
	# specify a different one.
    f <- list.files(pattern = "TXT")
    codset <- readCodelinkSet(filename = f)

	# read a targets file with phenotypic information
    # including sample names, file names and sample groups.
    pdata = read.AnnotationDataFrame("targets.txt")
    codset = readCodelinkSet(pdata$FileNames, phenoData=pdata)

    data(codset)
}
}
\keyword{documentation}
\keyword{utilities}

\name{arrayNew}
\alias{arrayNew}
\title{Create a new x11 device}
\description{
  Create a new x11 device with dimensions suited to be used
    with imageCodelink().
}
\usage{
  arrayNew(f=2, chip="rwgcod")
}
\arguments{
  \item{chip}{character; Codelink chip to be used in imageCodelink.}
  \item{f}{numerical; scaling factor.}
}
\author{Diego Diez}
\examples{
\dontrun{
    data(codelink.example)
    arrayNew()
    imageCodelink(codelink.example)
}
}
\keyword{documentation}
\keyword{utilities}

\name{Codelink-class}
\docType{class}
\alias{class:Codelink}
\alias{Codelink}
\alias{Codelink-class}
\alias{[,Codelink-method}
\alias{[,Codelink,ANY,ANY-method}
\alias{show,Codelink-method}

\title{Class Codelink}
\description{This is the storage class for Codelink data}
\section{Objects from the Class}{
  Object are created after reading text codelink files with readCodelink()}

\section{Description}{
	Codelink objects contain a single \code{"list"}, which contains the following elements:
  \describe{
    \item{sample}{Object of class \code{"character"} containing the sample names}
    \item{file}{Object of class \code{"character"} containing the file names}
	\item{name}{Object of class \code{"character"} containing the probe ID}
    \item{method}{Object of class \code{"list"} containing log information}
    \item{Smean}{Object of class \code{"matrix"} containing spot mean intensities}
    \item{Bmedian}{Object of class \code{"matrix"} containing background median intensities}
    \item{Ri}{Object of class \code{"matrix"} containing raw intensities}
    \item{Ni}{Object of class \code{"matrix"} containing normalized intensities}
    \item{snr}{Object of class \code{"matrix"} containing signal to noise ratio values}
    \item{flag}{Object of class \code{"character"} containing assigned flags}
  }
}

\note{More details are in the package vignette}

\author{Diego Diez}

\examples{
\dontrun{
    data(codelink.example)
}
}
\keyword{classes}

\name{createWeights}
\alias{createWeights}
\alias{createWeights,CodelinkSet-method}
\alias{createWeights,Codelink-method}
\title{Create weight for Codelink chips}
\description{
  Create a weight matrix based on probe type and flags.
}
\usage{
	createWeights(object, type.weights = NULL, flag.weights = NULL)
}
\arguments{
	\item{object}{an object of class "Codelink" or "CodelinkSet".}
	\item{type.weights}{named character vector of type weights used.}
	\item{flag.weights}{named character vector of flag weights used.}
}
\author{Diego Diez}
\examples{
    data(codset)
    w <- createWeights(codset)
}
\keyword{documentation}
\keyword{utilities}

\name{cutCV}
\alias{cutCV}
\title{Calculate cutoff based in CV.}
\description{
	Takes a Codelink object and calculate cutoff based in CV.
}
\usage{
	cutCV(object, subset=c(1:dim(object)[2]))
}
\arguments{
	\item{object}{an object of class "Codelink".}
	\item{subset}{subset of arrays to calculate cutoff with}
}
\details{
    First it computes the median of CV for each gene over all arrays. Then
  it computes the mean and sd of all medians. Finally:

    cutoff = mean + 3 * sd

}
\author{Diego Diez}
\examples{
\dontrun{
    # data: Normalized Codelink object merged.
    cutoff <- cutCV(data)
}
}
\keyword{documentation}
\keyword{utilities}

\name{readHeader}
\alias{readHeader}
\title{Read Header from Codelink Bioarrays Files}
\description{
    Read the header of Codelink files and obtein usefull information.
}
\usage{
readHeader(file, dec=FALSE)
}
\arguments{
  \item{file}{File to read.}
  \item{dec}{logical; If TRUE determine decimal point.}
}
\value{A list with header and other usefull information.}
\author{Diego Diez}
\examples{
\dontrun{
    files <- list.files(pattern = "TXT")
    head <- readHeader(files[1])
}
}
\keyword{documentation}
\keyword{utilities}

\name{averageProbes}
\alias{averageProbes}
\alias{averageProbes,CodelinkSet-method}
\title{averageProbes}
\description{
	Takes a CodelinkSet object and computes the average (mean) and sd of duplicated
	probes.
}
\usage{
	averageProbes(object, parallel = FALSE)
}
\arguments{
  \item{object}{an object of class "CodelinkSet".}
  \item{parallel}{whether to use a parallel version (requires package multicore).}
}
\details{

	This function will compute the mean() and sd() on each duplicated probe (i.e.
	identical probe id as for probeNames(). CodelinkSet objects use the per-array
	index to enable unique identifiers needed for ExpressionSet derived objects.
	Although the method probeNames() provides access to Codelink probe ids, this is
	inconvenient when dealing with other packages that make use of featureNames()
	to obtain probe ids and feed them  to the corresponding annotation package.

	In such cases CodelinkSet objects are not compatible with methods working on
	ExpressionSet objects. To avoid this limitation it is possible to construct a
	CodelinkSet object containing unique identifiers (CodelinkSetUnique class) by
	averaging the intensities of all replicated probes. This is done by computing
	the mean(). The sd() is also computed and stored in the slot sd.

	The current implementation takes a lot of time so a parallelized version of 
	lapply() may be used through the package 'multicore'. This is controlled by the
	argument 'parallel' which is FALSE by default.

	According to the authors it is not actually possible to use 'multicore' in a 
	GUI environment like the R.app Cocoa application in R, because this interferes 
	with the events loop. Therefore it is adviced to use option 'parallel=TRUE' in
	an R session running in a shell.

}
\author{Diego Diez}
\examples{
\dontrun{
    data(codelink.example)
	foo <- averageProbes(codelink.example)
}
}
\keyword{documentation}
\keyword{utilities}

\name{readHeaderXLS}
\alias{readHeaderXLS}
\title{Read Header from XLS exported Codelink Bioarrays Files}
\description{
    Read the header of Codelink files and obtain useful information.
}
\usage{
readHeaderXLS(file, dec=FALSE)
}
\arguments{
  \item{file}{File to read.}
  \item{dec}{logical; If TRUE determine decimal point.}
}
\value{A list with header and other usefull information.}
\details{
	This function is not meant to be used by normal users.
}
\author{Diego Diez}
\examples{
\dontrun{
    files <- list.files(pattern = "TXT")
    head <- readHeaderHeader(files[1])
}
}
\keyword{documentation}
\keyword{utilities}

\name{imageCodelink}
\alias{imageCodelink}
\title{Image plot of Codelink arrays}
\description{
  Plot and image of a Codelink array if the layout information is found.
}
\usage{

imageCodelink(object, array = 1, what = "bg",
    low="black", high="white", mar=c(1,1,1,1),
    gr=1, gc=1, log.it=FALSE, ...)
	
}
\arguments{
  \item{object}{an object of class "Codelink".}
  \item{array}{array to be used.}
  \item{what}{with data plot: bg, smean, ri, ni.}
  \item{low}{color used for low intensities.}
  \item{high}{color used for high intensities-}
  \item{mar}{character vector specifiying margings.}
  \item{gc}{numerical; number of grid columns.}
  \item{gr}{numerical; number of grid rows.}
  \item{log.it}{logical; if TRUE data is log2 transformed (if not yet).}
  \item{...}{additional arguments passed to image.}
}
\author{Diego Diez}
\examples{
\dontrun{
    data(codelink.example)
    imageCodelink(codelink.example)
}
}
\keyword{documentation}
\keyword{utilities}

\name{na2false}
\alias{na2false}
\title{Set NAs to FALSE}
\description{
	Takes a logical vector as input and set all NAs to FALSE. This may happens when comparison is done on NA values.
}
\usage{
	na2false(x)
}
\arguments{
  \item{x}{a logical vector.}
}
\value{A logical vector without NAs.}
\author{Diego Diez}
\examples{
\dontrun{
    a <- c(1, 2, 3, NA, 5)
    b <- c(5, 4, NA, 2, 1)
    sel <- a > b
    sel <- na2false(sel)
}
}
\keyword{documentation}
\keyword{utilities}

\name{mergeArray}
\alias{mergeArray}
\title{Merge Codelink Bioarrays Data}
\description{
  Merge data in a Codelink Object corresponding to same samples. Need a vector
indicating the classes and an optional vector indicating the labels of the mer-
ged samples.
}
\usage{
  mergeArray(object, class, names=NULL, method="mean",
             log.it=FALSE, merge.snr=TRUE)
}
\arguments{
  \item{object}{an object of class "Codelink".}
  \item{class}{a numeric vector indicating the classes.}
  \item{names}{an optional character vector indicating labels for each class.}
  \item{method}{the method used to summarize. Currently only "mean" supported.}
  \item{log.it}{logical; a logical indicating if log2 values should be returned.}
  \item{merge.snr}{logical; a logical indicating if SNR values should be merged.}
}
\value{An object of class "Codelink".}
\author{Diego Diez}
\examples{
\dontrun{
	data(codelink.example)
	codelink.example <- bkgdCorrect(codelink.example)
	codelink.example <- normalize(codelink.example, log.it = FALSE)
	codelink.example <- mergeArray(codelink.example, class = c(1,1),
					names = "SAMPLE", log.it = TRUE)
}
}
\keyword{documentation}
\keyword{utilities}

\name{dim.Codelink}
\alias{dim.Codelink}
\title{Return the dimension of a Codelink object.}
\description{
	Takes a Codelink object and returns the dimension (genes x samples).
}
\usage{
  \method{dim}{Codelink}(x)
}
\arguments{
  \item{x}{an object of class "Codelink".}
}
\value{A numeric vector with the dimensions.}
\author{Diego Diez}
\examples{
\dontrun{
	data(codelink.example)
	dim(codelink.example)
}
}
\keyword{documentation}
\keyword{utilities}

\name{printHead}
\alias{printHead}
\title{Print briefly a Codelink object}
\description{
	Takes a Codelink object and print a summary information of the data estored. It is based on printHead() from package limma.
}
\usage{
	printHead(x)
}
\arguments{
  \item{x}{an object of class "Codelink".}
}
\author{Diego Diez}
\keyword{documentation}
\keyword{utilities}

\name{CodelinkSetUnique-class}
\alias{class::CodelinkSetUnique}
\alias{CodelinkSetUnique}
\alias{CodelinkSetUnique-class}
\docType{class}

\title{Class CodelinkSetUnique}
\description{This is the storage class for Codelink data-- with unique probes}
\section{Objects from the Class}{
  Object are created after applying averageProbes() on an object of the class CodelinkSet
}

\section{Description}{
	CodelinkSetUnique objects are derived from ExpressionSet and therefore inherits all the methods.}

\note{More details are in the package vignette}

\author{Diego Diez}

\examples{
\dontrun{
    data(codelink.exprs)
    foo <- averageProbes(codelink.exprs)
}
}
\keyword{classes}

\name{selCV}
\alias{selCV}
\title{Select based on CV cutoff.}
\description{
	Takes a Codelink object and select genes based on CV cutoff.
}
\usage{
	selCV(object, cutoff)
}
\arguments{
	\item{object}{an object of class "Codelink".}
	\item{cutoff}{cutoff normaly calculated with cutCV()}
}
\value{
	A logical vector.
}
\author{Diego Diez}
\keyword{documentation}
\keyword{utilities}
